# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# ruff: noqa: S105, S106

import copy
import time
from datetime import datetime, timezone
from math import ceil

import httpx
import pytest
from pytest_httpx import HTTPXMock

from ansys.simai.core.errors import SimAIError
from ansys.simai.core.utils.auth import (
    TOKEN_REFRESH_BUFFER,
    Authenticator,
    _AuthTokens,
    _AuthTokensRetriever,
)
from ansys.simai.core.utils.configuration import ClientConfig, Credentials

DEFAULT_TOKENS = {
    "access_token": "check",
    "expires_in": 60,
    "refresh_expires_in": 1800,
    "refresh_token": "kaboom",
}


def test_request_auth_tokens_direct_grant_bad_credentials_raises(mocker, tmpdir, httpx_mock):
    mocker.patch("ansys.simai.core.utils.auth.get_cache_dir", return_value=tmpdir)
    httpx_mock.add_response(
        method="POST",
        url="http://myauthserver.com/protocol/openid-connect/token",
        json={
            "error": "invalid_grant",
            "error_description": "Invalid user credentials",
        },
        status_code=401,
    )
    tokens_retriever = _AuthTokensRetriever(
        credentials=None,
        session=httpx.Client(),
        realm_url="http://myauthserver.com",
        auth_cache_hash="rando",
    )
    with pytest.raises(SimAIError, match="Invalid user credentials"):
        tokens_retriever.credentials = Credentials(username="macron", password="explosion")
        tokens_retriever.get_tokens()


def test_request_auth_tokens_direct_grant(mocker, tmpdir, httpx_mock):
    mocker.patch("ansys.simai.core.utils.auth.get_cache_dir", return_value=tmpdir)
    httpx_mock.add_response(
        method="POST",
        url="http://myauthserver.com/protocol/openid-connect/token",
        json=DEFAULT_TOKENS,
        status_code=200,
    )
    tokens_retriever = _AuthTokensRetriever(
        credentials=None,
        session=httpx.Client(),
        realm_url="http://myauthserver.com",
        auth_cache_hash="rando",
    )
    tokens_retriever.credentials = Credentials(username="timmy", password="")
    tokens = tokens_retriever.get_tokens()
    assert tokens.refresh_token == DEFAULT_TOKENS["refresh_token"]
    assert tokens.access_token == DEFAULT_TOKENS["access_token"]


def test_token_refresh_failure_triggers_reauth(mocker, tmpdir, httpx_mock):
    mocker.patch("ansys.simai.core.utils.auth.get_cache_dir", return_value=tmpdir)
    httpx_mock.add_response(
        method="POST",
        url="http://myauthserver.com/protocol/openid-connect/token",
        status_code=418,
    )
    httpx_mock.add_response(
        method="POST",
        url="http://myauthserver.com/protocol/openid-connect/token",
        json=DEFAULT_TOKENS,
        status_code=200,
    )
    with open(tmpdir / "tokens-rando.json", "w") as f:
        f.write(
            _AuthTokens(
                access_token="",
                expires_in=0,
                refresh_expires_in=999,
                refresh_token="revoked",
            ).model_dump_json()
        )
    tokens_retriever = _AuthTokensRetriever(
        credentials=Credentials(username="timmy", password=""),
        session=httpx.Client(),
        realm_url="http://myauthserver.com",
        auth_cache_hash="rando",
    )
    tokens = tokens_retriever.get_tokens()
    assert tokens.refresh_token == DEFAULT_TOKENS["refresh_token"]
    assert tokens.access_token == DEFAULT_TOKENS["access_token"]
    # Note: With pytest-httpx, we can't easily check call counts like with responses
    # The test verifies that the tokens are retrieved correctly, which is the main goal


def test_request_auth_tokens_device_grant_with_bad_cache(mocker, tmpdir, httpx_mock):
    """WHEN A device auth flow is requested
    AND a (bad) refresh token is cached
    THEN SDK tries to use the cached refresh token
    AND requests a device auth flow if the token is invalid
    """
    webbrowser_open = mocker.patch("webbrowser.open")
    mocker.patch("ansys.simai.core.utils.auth.get_cache_dir", return_value=tmpdir)
    mocker.patch("ansys.simai.core.utils.auth.DEVICE_AUTH_POLLING_INTERVAL", 0)
    refresh_token = "spaghetti"
    device_code = "foxtrot uniform charlie kilo"
    realm_url = "http://myauthserver.com/my-realm"
    token_retriever = _AuthTokensRetriever(
        credentials=None, session=httpx.Client(), realm_url=realm_url, auth_cache_hash="lol"
    )
    fake_cache = _AuthTokens(
        access_token="",
        expires_in=0,
        refresh_expires_in=999,
        refresh_token=refresh_token,
    )
    with open(tmpdir / "tokens-lol.json", "w") as f:
        f.write(fake_cache.model_dump_json())

    httpx_mock.add_response(  # error when app tries to use invalidated cached token
        method="POST",
        url=token_retriever.token_url,
        json={"poop": "Ur token down the drain"},
        status_code=418,
    )
    httpx_mock.add_response(  # device-auth flow start
        method="POST",
        url=token_retriever.device_auth_url,
        json={
            "verification_uri": "nope",
            "user_code": "abcdefg",
            "verification_uri_complete": "nope?code=abcdefg",
            "device_code": device_code,
        },
        status_code=200,
    )
    httpx_mock.add_response(  # device-auth flow end
        method="POST",
        url=token_retriever.token_url,
        json=DEFAULT_TOKENS,
        status_code=200,
    )

    tokens = token_retriever.get_tokens()
    webbrowser_open.assert_called_with("nope?code=abcdefg")
    assert tokens.access_token == DEFAULT_TOKENS["access_token"]
    assert tokens.refresh_token == DEFAULT_TOKENS["refresh_token"]
    assert ceil(tokens.expires_in) == DEFAULT_TOKENS["expires_in"]
    assert ceil(tokens.refresh_expires_in) == DEFAULT_TOKENS["refresh_expires_in"]


def test_refresh_auth_tokens(mocker, tmpdir, httpx_mock):
    mocker.patch("ansys.simai.core.utils.auth.get_cache_dir", return_value=tmpdir)
    base_token = copy.deepcopy(DEFAULT_TOKENS)
    base_token["refresh_token"] = "kabam"
    final_token = copy.deepcopy(base_token)
    final_token["access_token"] = "new tok"

    expired_token = _AuthTokens.model_validate(copy.deepcopy(base_token))
    expired_token.expiration = datetime(year=1970, month=1, day=1, tzinfo=timezone.utc)
    token_retriever = _AuthTokensRetriever(
        credentials=None,
        session=httpx.Client(),
        realm_url="http://myauthserver.com",
        auth_cache_hash="popo",
    )
    token_retriever._get_token_from_cache = lambda: expired_token

    httpx_mock.add_response(
        method="POST",
        url="http://myauthserver.com/protocol/openid-connect/token",
        json=final_token,
        status_code=200,
    )

    fetched_token = token_retriever.get_tokens()
    assert fetched_token.access_token == final_token["access_token"]
    assert fetched_token.refresh_token == final_token["refresh_token"]


def test_authenticator_automatically_refreshes_auth_before_requests_if_needed(
    mocker, tmpdir, httpx_mock
):
    # Authentication request
    auth_tokens = copy.deepcopy(DEFAULT_TOKENS)
    auth_tokens["expires_in"] = 1
    auth_tokens["refresh_expires_in"] = 1800
    httpx_mock.add_response(
        method="POST",
        url="https://simai.ansys.com/auth/realms/simai/protocol/openid-connect/token",
        json=auth_tokens,
        status_code=200,
    )
    # Token refresh request
    refresh_tokens = copy.deepcopy(DEFAULT_TOKENS)
    refresh_tokens["access_token"] = "check 1 2"
    refresh_tokens["refresh_token"] = "megazaur"
    httpx_mock.add_response(
        method="POST",
        url="https://simai.ansys.com/auth/realms/simai/protocol/openid-connect/token",
        json=refresh_tokens,
        status_code=200,
    )

    mocker.patch("ansys.simai.core.utils.auth.get_cache_dir", return_value=tmpdir)
    auth = Authenticator(
        ClientConfig(
            url="https://simai.ansys.com",
            organization="13_monkeys",
            credentials=Credentials(username="timmy", password="key"),
        ),
        httpx.Client(),
    )

    req = httpx.Request("GET", "https://simai.ansys.com/v2/models")
    req = next(auth.auth_flow(req))
    assert req.headers.get("Authorization") == "Bearer check 1 2"
    assert req.headers.get("X-Org") == "13_monkeys"


def test_authenticator_automatically_refreshes_auth_before_refresh_token_expires(
    mocker, tmpdir, httpx_mock: HTTPXMock
):
    """
    Test that the Authenticator schedules and performs an automatic token refresh
    before the refresh token expires, based on the TOKEN_REFRESH_BUFFER.

    Verifies that when a refresh token is still valid but close to expiration,
    the authenticator schedules a refresh operation to prevent token expiration
    during idle periods.
    """
    mocker.patch("ansys.simai.core.utils.auth.get_cache_dir", return_value=tmpdir)

    tokens_direct_grant = copy.deepcopy(DEFAULT_TOKENS)
    tokens_direct_grant["access_token"] = "monkey-see"
    tokens_direct_grant["expires_in"] = TOKEN_REFRESH_BUFFER + 100
    tokens_direct_grant["refresh_expires_in"] = TOKEN_REFRESH_BUFFER + 1
    httpx_mock.add_response(
        method="POST",
        url="https://simai.ansys.com/auth/realms/simai/protocol/openid-connect/token",
        json=tokens_direct_grant,
        status_code=200,
    )
    matcher_direct_grant = httpx_mock._callbacks[-1][0]

    refreshed_tokens = DEFAULT_TOKENS.copy()
    refreshed_tokens["access_token"] = "TFou"
    httpx_mock.add_response(
        method="POST",
        url="https://simai.ansys.com/auth/realms/simai/protocol/openid-connect/token",
        json=refreshed_tokens,
        status_code=200,
    )
    matcher_refresh = httpx_mock._callbacks[-1][0]
    auth = Authenticator(
        ClientConfig(
            url="https://simai.ansys.com",
            organization="14_monkeys",
            credentials=Credentials(username="timmy", password="key"),
        ),
        httpx.Client(),
    )
    initial_refresh_timer = auth.tokens_retriever.refresh_timer
    assert matcher_direct_grant.nb_calls == 1
    assert matcher_refresh.nb_calls == 0
    assert initial_refresh_timer.is_alive()
    t0 = time.time()
    while time.time() - t0 < 2 and auth.tokens_retriever.refresh_timer == initial_refresh_timer:
        # wait for the refresh timer to restart, or for the 2sec timeout...
        time.sleep(0.1)
    assert matcher_direct_grant.nb_calls == 1  # Initial authentication
    assert matcher_direct_grant.nb_calls == 1  # Tokens refresh
    assert auth.tokens_retriever.refresh_timer != initial_refresh_timer
    assert auth.tokens_retriever.refresh_timer.is_alive()


def test_requests_outside_user_api_are_not_authentified(mocker, tmpdir, httpx_mock):
    mocker.patch("ansys.simai.core.utils.auth.get_cache_dir", return_value=tmpdir)
    # Authentication request
    keycloak_response_json = DEFAULT_TOKENS.copy()
    httpx_mock.add_response(
        method="POST",
        url="https://simai.ansys.com/auth/realms/simai/protocol/openid-connect/token",
        json=keycloak_response_json,
        status_code=200,
    )

    auth = Authenticator(
        ClientConfig(
            url="https://simai.ansys.com",
            organization="Justice",
            credentials={"username": "timmy", "password": "D.A.N.C.E"},
        ),
        httpx.Client(),
    )

    request = httpx.Request("GET", "https://amazonaws.com/bloc-party")
    request = next(auth.auth_flow(request))
    assert request.headers.get("Authorization") is None
    assert request.headers.get("X-Org") is None


def test_get_offline_token_direct_grant(httpx_mock):
    offline_tokens = {
        "access_token": "access",
        "expires_in": 300,
        "refresh_expires_in": 0,
        "refresh_token": "offline-refresh-token",
    }
    httpx_mock.add_response(
        method="POST",
        url="https://simai.ansys.com/auth/realms/simai/protocol/openid-connect/token",
        json=offline_tokens,
        status_code=200,
    )
    from ansys.simai.core.utils.auth import get_offline_token

    token = get_offline_token(
        url="https://simai.ansys.com",
        credentials=Credentials(username="user", password="pass"),
    )
    assert token == "offline-refresh-token"
    request = httpx_mock.get_request()
    assert b"scope=openid+offline_access" in request.content


def test_auth_with_offline_token(mocker, tmpdir, httpx_mock):
    """WHEN authenticating with an offline token
    THEN the offline token is used to get access tokens via refresh grant.
    """
    mocker.patch("ansys.simai.core.utils.auth.get_cache_dir", return_value=tmpdir)

    tokens_with_long_expiry = copy.deepcopy(DEFAULT_TOKENS)
    tokens_with_long_expiry["expires_in"] = 3600
    tokens_with_long_expiry["refresh_expires_in"] = 7200
    httpx_mock.add_response(
        method="POST",
        url="https://simai.ansys.com/auth/realms/simai/protocol/openid-connect/token",
        json=tokens_with_long_expiry,
        status_code=200,
    )

    auth = Authenticator(
        ClientConfig(
            url="https://simai.ansys.com",
            organization="test_org",
            offline_token="my-offline-token",
        ),
        httpx.Client(),
    )

    request = httpx_mock.get_request()
    assert b"grant_type=refresh_token" in request.content
    assert b"refresh_token=my-offline-token" in request.content

    req = httpx.Request("GET", "https://simai.ansys.com/v2/models")
    req = next(auth.auth_flow(req))
    assert req.headers.get("Authorization") == f"Bearer {tokens_with_long_expiry['access_token']}"
    assert req.headers.get("X-Org") == "test_org"


def test_offline_token_and_credentials_mutually_exclusive():
    """WHEN both credentials and offline_token are provided
    THEN a validation error is raised.
    """
    from pydantic import ValidationError

    with pytest.raises(ValidationError, match="auth_conflict"):
        ClientConfig(
            url="https://simai.ansys.com",
            organization="test_org",
            credentials=Credentials(username="user", password="pass"),
            offline_token="my-offline-token",
        )


def test_non_interactive_mode_accepts_offline_token():
    """WHEN non-interactive mode is used with offline_token
    THEN the config is valid (no credentials required).
    """
    config = ClientConfig(
        url="https://simai.ansys.com",
        organization="test_org",
        interactive=False,
        offline_token="my-offline-token",
    )
    assert config.offline_token == "my-offline-token"
    assert config.credentials is None



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest


@pytest.fixture()
def geometries_with_two_metadata(create_mock_geometry):
    """A list of geometries with 2 numerical metadata:
    - 1 int metadata (length), range 0 to 6
    - 1 float metadata (width), range 0.0 to 0.6
    """
    geometries = []
    for i in range(7):
        for j in range(7):
            geometries.append(create_mock_geometry(id=f"l-{i}-w-{j}", length=i, width=j / 10))
    yield geometries


@pytest.fixture()
def geometries_with_two_metadata_and_incomplete_ones(create_mock_geometry):
    """A list of geometries with 2 numerical metadata:
    - 1 int metadata (length), range 0 to 6
    - 1 float metadata (width), range 0.0 to 0.6
    """
    geometries = []
    for i in range(7):
        for j in range(7):
            geometries.append(create_mock_geometry(id=f"l-{i}-w-{j}", length=i, width=j / 10))
    # create a double of 2-2
    geometries.append(create_mock_geometry(id="l-2-w-2-bis", length=2, width=0.2))
    # add incomplete geometries
    geometries.append(create_mock_geometry(id="l-3", length=3))
    geometries.append(create_mock_geometry(id="w-5", width=0.5))
    yield geometries


def create_length_width_set(geometries):
    return {(g.metadata["length"], g.metadata["width"]) for g in geometries}


def test_sweep_2_dimensions_center(geometry_directory, geometries_with_two_metadata):
    """WHEN running a sweep around a (2, 0.3) geometry,
    THEN 4 geometries are returned: (1, 0.3), (3, 0.3), (2, 0.2), (2, 0.4)
    IF including diagonals,
    THEN 4 additional corner geometries are returned: (1, 0.2), (1, 0.4), (3, 0.2), (3, 0.4)
    IF including center
    THEN the candidate (2, 0.3) is included in the results
    """

    candidate = geometries_with_two_metadata[17]
    assert candidate.metadata["length"] == 2
    assert candidate.metadata["width"] == 0.3

    direct_neighbors = {(1, 0.3), (3, 0.3), (2, 0.2), (2, 0.4)}
    all_neighbors = direct_neighbors.union({(1, 0.2), (1, 0.4), (3, 0.2), (3, 0.4)})

    for include_diagonals in [False, True]:
        for include_center in [False, True]:
            swept = geometry_directory.sweep(
                candidate_geometry=candidate,
                geometries=geometries_with_two_metadata,
                include_center=include_center,
                include_diagonals=include_diagonals,
            )
            length_widths = create_length_width_set(swept)
            expected_neighbors = all_neighbors if include_diagonals else direct_neighbors
            if include_center:
                expected_neighbors.add((2, 0.3))
            assert length_widths == expected_neighbors


def test_sweep_2_dimensions_center_order_2(geometry_directory, geometries_with_two_metadata):
    """WHEN running a sweep around a (2, 0.3) geometry with order 2
    THEN 8 geometries are returned:
        (0, 0.3), (1, 0.3), (3, 0.3), (4, 0.3),
        (2, 0.1), (2, 0.2), (2, 0.4), (2, 0.5)
    IF including center
    THEN the candidate (2, 0.3) is included in the results
    """

    candidate = geometries_with_two_metadata[17]
    assert candidate.metadata["length"] == 2
    assert candidate.metadata["width"] == 0.3

    direct_neighbors = {
        (0, 0.3),
        (1, 0.3),
        (3, 0.3),
        (4, 0.3),
        (2, 0.1),
        (2, 0.2),
        (2, 0.4),
        (2, 0.5),
    }

    for include_center in [False, True]:
        swept = geometry_directory.sweep(
            candidate_geometry=candidate,
            geometries=geometries_with_two_metadata,
            order=2,
            include_center=include_center,
        )
        length_widths = create_length_width_set(swept)
        expected_neighbors = direct_neighbors
        if include_center:
            expected_neighbors.add((2, 0.3))
        assert length_widths == expected_neighbors


def test_sweep_2_dimensions_center_order_2_with_diagonals(
    geometry_directory, geometries_with_two_metadata
):
    """WHEN running a sweep around a (2, 0.3) geometry with order 2 including diagonals
    THEN all those geometries are returned:
        (0, 0.1), (1, 0.1), (2, 0.1), (3, 0.1), (4, 0.1)
        (0, 0.2), (1, 0.2), (2, 0.2), (3, 0.2), (4, 0.2)
        (0, 0.3), (1, 0.3),           (3, 0.3), (4, 0.3)
        (0, 0.4), (1, 0.4), (2, 0.4), (3, 0.4), (4, 0.4)
        (0, 0.5), (1, 0.5), (2, 0.5), (3, 0.5), (4, 0.5)
    IF including center
    THEN the candidate (2, 0.3) is included in the results
    """

    candidate = geometries_with_two_metadata[17]
    assert candidate.metadata["length"] == 2
    assert candidate.metadata["width"] == 0.3

    neighbors = {
        (0, 0.1),
        (1, 0.1),
        (2, 0.1),
        (3, 0.1),
        (4, 0.1),
        (0, 0.2),
        (1, 0.2),
        (2, 0.2),
        (3, 0.2),
        (4, 0.2),
        (0, 0.3),
        (1, 0.3),
        (3, 0.3),
        (4, 0.3),
        (0, 0.4),
        (1, 0.4),
        (2, 0.4),
        (3, 0.4),
        (4, 0.4),
        (0, 0.5),
        (1, 0.5),
        (2, 0.5),
        (3, 0.5),
        (4, 0.5),
    }

    for include_center in [False, True]:
        swept = geometry_directory.sweep(
            candidate_geometry=candidate,
            geometries=geometries_with_two_metadata,
            order=2,
            include_center=include_center,
            include_diagonals=True,
        )
        length_widths = create_length_width_set(swept)
        expected_neighbors = neighbors
        if include_center:
            expected_neighbors.add((2, 0.3))
        assert length_widths == neighbors


def test_sweep_2_dimensions_min_corner(geometry_directory, geometries_with_two_metadata):
    """WHEN running a sweep around a "min corner" (0, 0.0) geometry,
    THEN 2 geometries are returned: (1, 0) and (0, 0.1)
    IF including center
    THEN the candidate is included
    IF including diagonals
    THEN the diagonal geometry is returned: (1, 0.1)
    """
    candidate = geometries_with_two_metadata[0]
    assert candidate.metadata["length"] == 0
    assert candidate.metadata["width"] == 0

    for include_diagonals in [False, True]:
        for include_center in [False, True]:
            swept = geometry_directory.sweep(
                candidate_geometry=candidate,
                geometries=geometries_with_two_metadata,
                include_center=include_center,
                include_diagonals=include_diagonals,
            )
            length_widths = create_length_width_set(swept)
            expected_neighbors = {(1, 0), (0, 0.1)}
            if include_center:
                expected_neighbors.add((0, 0.0))
            if include_diagonals:
                expected_neighbors.add((1, 0.1))
            assert length_widths == expected_neighbors


def test_sweep_2_dimensions_max_corner(geometry_directory, geometries_with_two_metadata):
    """WHEN running a sweep around a "max corner" (6, 0.6) geometry,
    THEN 2 geometries are returned: (5, 0.6) and (6, 0.5)
    IF including center
    THEN the candidate is included
    IF include diagonals
    THEN (5, 0.5) is also returned
    """
    candidate = geometries_with_two_metadata[48]
    assert candidate.metadata["length"] == 6
    assert candidate.metadata["width"] == 0.6

    for include_diagonals in [False, True]:
        for include_center in [False, True]:
            swept = geometry_directory.sweep(
                candidate_geometry=candidate,
                geometries=geometries_with_two_metadata,
                include_center=include_center,
                include_diagonals=include_diagonals,
            )
            length_widths = create_length_width_set(swept)

            expected_neighbors = {(6, 0.5), (5, 0.6)}
            if include_center:
                expected_neighbors.add((6, 0.6))
            if include_diagonals:
                expected_neighbors.add((5, 0.5))
            assert length_widths == expected_neighbors


def test_sweep_2_dimensions_length_border(geometry_directory, geometries_with_two_metadata):
    """WHEN running a sweep around a "border" (0, 0.3) geometry,
    THEN 3 geometries are returned: (1, 0.3), (0, 0.2), (0, 0.4)
    """
    candidate = geometries_with_two_metadata[3]
    assert candidate.metadata["length"] == 0
    assert candidate.metadata["width"] == 0.3
    swept = geometry_directory.sweep(
        candidate_geometry=candidate, geometries=geometries_with_two_metadata
    )
    length_widths = create_length_width_set(swept)

    assert length_widths == {(1, 0.3), (0, 0.2), (0, 0.4)}


def test_sweep_2_dimensions_width_border(geometry_directory, geometries_with_two_metadata):
    """WHEN running a sweep around a "border" (4, 0.6) geometry,
    THEN 3 geometries are returned: (3, 0.6), (5, 0.6), (4, 0.5)
    """
    candidate = geometries_with_two_metadata[34]
    assert candidate.metadata["length"] == 4
    assert candidate.metadata["width"] == 0.6
    swept = geometry_directory.sweep(
        candidate_geometry=candidate, geometries=geometries_with_two_metadata
    )
    length_widths = create_length_width_set(swept)
    assert length_widths == {(3, 0.6), (5, 0.6), (4, 0.5)}


def test_sweep_2_dimensions_ignoring_length(geometry_directory, geometries_with_two_metadata):
    """WHEN running a sweep around a (3, 0.3) geometry along the second (width) metadata,
    THEN 14 geometries are returned:
    - all 7 geometries with width=0.2 (whatever length), and
    - all 7 geometries with width=0.4 (whatever length).
    ALSO if including center
    THEN results include all 7 geometries with width=0.3 (whatever length)
    ALSO including diagonals has no effect as we sweep only 1 variable
    """
    candidate = geometries_with_two_metadata[24]
    assert candidate.metadata["length"] == 3
    assert candidate.metadata["width"] == 0.3

    for include_diagonals in [False, True]:
        for include_center in [False, True]:
            swept = geometry_directory.sweep(
                candidate_geometry=candidate,
                swept_metadata=["width"],
                geometries=geometries_with_two_metadata,
                include_center=include_center,
                include_diagonals=include_diagonals,
            )
            assert len(swept) == (21 if include_center else 14)
            length_widths = create_length_width_set(swept)
            expected_results = set()
            for w in [0.2, 0.3, 0.4] if include_center else [0.2, 0.4]:
                for l in range(7):  # noqa: E741
                    expected_results.add((l, w))
            assert length_widths == expected_results


def test_sweep_2_dimensions_fixing_length(geometry_directory, geometries_with_two_metadata):
    """WHEN running a sweep around a (3, 0.3) geometry along the second (width) metadata,
      and fixing the length metadata
    THEN 2 geometries are returned:
    - geometry with width=0.2 and length=3, and
    - geometry with width=0.4 and length=3
    IF including center, then the candidate is returned
    ALSO including diagonals has no effect as we sweep only 1 variable
    """
    candidate = geometries_with_two_metadata[24]
    assert candidate.metadata["length"] == 3
    assert candidate.metadata["width"] == 0.3

    for include_diagonals in [False, True]:
        for include_center in [False, True]:
            swept = geometry_directory.sweep(
                candidate_geometry=candidate,
                swept_metadata=["width"],
                fixed_metadata=["length"],
                geometries=geometries_with_two_metadata,
                include_center=include_center,
                include_diagonals=include_diagonals,
            )
            length_widths = create_length_width_set(swept)
            if include_center:
                assert length_widths == {(3, 0.2), (3, 0.3), (3, 0.4)}
            else:
                assert length_widths == {(3, 0.2), (3, 0.4)}


def test_sweep_2_dimensions_center_big_order_including_diagonals(
    geometry_directory, geometries_with_two_metadata
):
    """WHEN running a sweep around a (2, 0.3) geometry with very large order and including diagonals
    THEN all geometries are returned, candidate only if include_center
    """

    candidate = geometries_with_two_metadata[17]
    assert candidate.metadata["length"] == 2
    assert candidate.metadata["width"] == 0.3

    set_of_all = create_length_width_set(geometries_with_two_metadata)

    for include_center in [False, True]:
        swept = geometry_directory.sweep(
            candidate_geometry=candidate,
            geometries=geometries_with_two_metadata,
            order=20,
            include_center=include_center,
            include_diagonals=True,
        )
        length_widths = create_length_width_set(swept)

        if include_center:
            assert length_widths == set_of_all
        else:
            assert length_widths == set_of_all - {(2, 0.3)}



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import TYPE_CHECKING, List, Optional, Union

from ansys.simai.core.data.base import DataModel, Directory
from ansys.simai.core.data.geomai.models import GeomAIModel, GeomAIModelConfiguration
from ansys.simai.core.data.types import Identifiable, get_id_from_identifiable
from ansys.simai.core.errors import InvalidArguments, ProcessingError

if TYPE_CHECKING:
    from ansys.simai.core.data.geomai.training_data import GeomAITrainingData
    from ansys.simai.core.data.geomai.workspaces import GeomAIWorkspace


class GeomAIProject(DataModel):
    """Provides the local representation of a GeomAI project object."""

    def __repr__(self) -> str:
        return f"<GeomAIProject: {self.id}, {self.name}>"

    @property
    def name(self) -> str:
        """Name of project."""
        return self.fields["name"]

    def rename(self, new_name: str) -> None:
        """Rename the project.

        Args:
            new_name: New name to give to the project.
        """
        self._client._api.update_geomai_project(self.id, name=new_name)
        self.reload()

    def list_training_data(self) -> List["GeomAITrainingData"]:
        """Lists all :class:`~.training_data.GeomAITrainingData` instances in the project."""
        raw_td_list = self._client._api.iter_training_data_in_geomai_project(self.id)
        return [
            self._client.geomai.training_data._model_from(training_data)
            for training_data in raw_td_list
        ]

    def data(self) -> List["GeomAITrainingData"]:
        """(**Deprecated**) Lists all :class:`~.training_data.GeomAITrainingData` instances in the project.

        Use :py:meth:`.list_training_data` instead.
        """
        return self.list_training_data()

    def list_workspaces(self) -> List["GeomAIWorkspace"]:
        """Lists all :class:`~.workspaces.GeomAIWorkspace` instances in the project."""
        workspaces = self._client._api.get_geomai_project_related_workspaces(self.id)
        return [self._client.geomai.workspaces._model_from(workspace) for workspace in workspaces]

    def workspaces(self) -> List["GeomAIWorkspace"]:
        """(**Deprecated**) Lists all :class:`~.workspaces.GeomAIWorkspace` instances in the project.

        Use :py:meth:`.list_workspaces` instead.
        """
        return self.list_workspaces()

    def list_models(self) -> List[GeomAIModel]:
        """Lists all :class:`~.models.GeomAIModel` instances in the project."""
        raw_model_list = self._client._api.get_geomai_project_models(self.id)
        return [self._client.geomai.models._model_from(model) for model in raw_model_list]

    @property
    def last_model_configuration(self) -> Optional[GeomAIModelConfiguration]:
        """Last :class:`configuration <.models.GeomAIModelConfiguration>` used for model training in this project."""
        raw_last_model_configuration = self.fields.get("last_model_configuration")
        if raw_last_model_configuration is None:
            return None
        return GeomAIModelConfiguration(**raw_last_model_configuration)

    def delete(self) -> None:
        """Delete the project."""
        self._client._api.delete_geomai_project(self.id)

    def cancel_build(self):
        """Cancels a build if there is one pending.

        Raises:
            ProcessingError: If there is no build to cancel
        """

        self.reload()
        if self.fields.get("is_being_trained") is False:
            raise ProcessingError("No build pending for this project.")
        self._client._api.cancel_geomai_build(self.id)

    def build_model(self, configuration: Union[dict, GeomAIModelConfiguration]) -> GeomAIModel:
        """Launches a GeomAI build with the given configuration.

        Args:
            configuration: the configuration to run the model with.
                See :class:`.models.GeomAIModelConfiguration` for details.
        """
        return self._client.geomai.models.build(self.id, configuration)

    def create_workspace(self, name: str) -> "GeomAIWorkspace":
        """Creates a workspace using the latest model trained in this project.

        Args:
            name: Name to give to the new workspace
        """
        return self._client.geomai.workspaces._model_from(
            self._client._api.create_geomai_workspace(name, self.id)
        )

    def set_as_current_project(self) -> None:
        """Configure the client to use this project instead of the one currently configured."""
        self._client.geomai.current_project = self


class GeomAIProjectDirectory(Directory[GeomAIProject]):
    """Provides a collection of methods related to projects.

    This class is accessed through ``client.projects``.

    Example:
        List all projects::

            import ansys.simai.core as asc

            simai_client = asc.from_config()
            simai_client.geomai.projects.list()
    """

    _data_model = GeomAIProject

    def list(self) -> List[GeomAIProject]:
        """List all projects available on the server."""
        return [self._model_from(data) for data in self._client._api.geomai_projects()]

    def create(self, name: str) -> GeomAIProject:
        """Create a project."""
        return self._model_from(self._client._api.create_geomai_project(name=name))

    def get(self, id: Optional[str] = None, name: Optional[str] = None) -> GeomAIProject:
        """Get a project by either ID or name.

        You can specify either the ID or the name, not both.

        Args:
            id: ID of the project.
            name: Name of the project.

        Returns:
            :class:`GeomAIProject` instance with the given ID if it exists.

        Raises:
            NotFoundError: If the project doesn't exist
        """
        if name and id:
            raise InvalidArguments("Cannot specify both 'id' and 'name' arguments.")
        elif name:
            return self._model_from(self._client._api.get_geomai_project_by_name(name))
        elif id:
            return self._model_from(self._client._api.get_geomai_project(id))
        else:
            raise InvalidArguments("Either 'id' or 'name' argument must be specified.")

    def delete(self, project: Identifiable[GeomAIProject]) -> None:
        """Delete a project.

        Args:
            project: ID or :class:`model <GeomAIProject>` of the project.
        """
        self._client._api.delete_geomai_project(get_id_from_identifiable(project))

    def cancel_build(self, project: Identifiable[GeomAIProject]):
        """Cancel a build if one is in progress.

        Args:
            project: ID or :class:`model <GeomAIProject>` of the project.
        """

        project_id = get_id_from_identifiable(project)
        project_response = self._client._api.get_geomai_project(project_id)
        if project_response.get("is_being_trained") is False:
            raise ProcessingError("No build pending for this project.")
        self._client._api.cancel_geomai_build(project_id)



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import TYPE_CHECKING

from ansys.simai.core.data.geomai.models import GeomAIModelConfiguration

if TYPE_CHECKING:
    from ansys.simai.core.data.geomai.models import GeomAIModel
    from ansys.simai.core.data.geomai.projects import GeomAIProject

NB_EPOCHS = 50
NB_LATENT_PARAMS = 10

MODEL_CONF_RAW = {
    "nb_epochs": NB_EPOCHS,
    "nb_latent_param": NB_LATENT_PARAMS,
}

MODEL_RAW = {
    "configuration": MODEL_CONF_RAW,
    "coreml_version": "5.666.333",
    "creation_time": "2035-07-08T16:26:51.943312",
    "error": None,
    "hidden": False,
    "id": "2yg3vpq3",
    "is_deployed": False,
    "name": "pspsps",
    "project_id": "r291h3yk",
    "state": "pending request",
    "updated_time": "2036-12-08T16:26:51.953502",
    "version": None,
}


def test_build(mocker, simai_client, httpx_mock):
    """WHEN I call build() with a working GeomAIModelConfiguration
    THEN I get a GeomAIModel object, its project_id matches the
    id of the project, and its configuration is a
    GeomAIModelConfiguration and its content matches the raw conf.
    """
    httpx_mock.add_response(
        method="POST",
        url=f"https://test.test/geomai/projects/{MODEL_RAW['project_id']}/models",
        json=MODEL_RAW,
        status_code=200,
    )

    raw_project = {
        "id": MODEL_RAW["project_id"],
        "name": "fifi",
    }

    project: GeomAIProject = simai_client.geomai._project_directory._model_from(raw_project)

    model_config = GeomAIModelConfiguration(**MODEL_CONF_RAW)
    launched_model: GeomAIModel = simai_client.geomai.models.build(project, model_config)

    assert isinstance(launched_model.configuration, GeomAIModelConfiguration)
    assert launched_model.project_id == MODEL_RAW["project_id"]

    assert launched_model.configuration == GeomAIModelConfiguration(**MODEL_CONF_RAW)


def test_build_with_config_as_dict(mocker, simai_client, httpx_mock):
    """WHEN I call build() with a working dict as model configuration
    THEN I get a GeomAIModel object, its project_id matches the
    id of the project, and its configuration is a
    GeomAIModelConfiguration and its content matches the raw conf.
    """
    httpx_mock.add_response(
        method="POST",
        url=f"https://test.test/geomai/projects/{MODEL_RAW['project_id']}/models",
        json=MODEL_RAW,
        status_code=200,
    )

    raw_project = {
        "id": MODEL_RAW["project_id"],
        "name": "fifi",
    }

    project: GeomAIProject = simai_client.geomai._project_directory._model_from(raw_project)

    launched_model: GeomAIModel = simai_client.geomai.models.build(project, MODEL_CONF_RAW)

    assert isinstance(launched_model.configuration, GeomAIModelConfiguration)
    assert launched_model.project_id == MODEL_RAW["project_id"]

    assert launched_model.configuration == GeomAIModelConfiguration(**MODEL_CONF_RAW)



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
from typing import TYPE_CHECKING
from urllib.parse import urlencode

import pytest

from ansys.simai.core.data.types import SubsetEnum
from ansys.simai.core.errors import InvalidArguments

if TYPE_CHECKING:
    from ansys.simai.core.data.training_data import TrainingData


def test_training_data_iter(simai_client, httpx_mock):
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/training-data",
        headers={
            "X-Pagination": json.dumps({"total": 999}),
        },
        json=[{"id": "one"}],
        status_code=200,
    )
    it = simai_client.training_data.iter()
    assert len(it) == 999
    assert next(it).id == "one"
    assert len(it) == 998
    assert next(it, None) is None
    assert len(it) == 998


def test_training_data_list(simai_client, httpx_mock):
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/training-data",
        headers={
            "X-Pagination": json.dumps({"total_pages": 2}),
            "Link": '<https://test.test/training-data?last_id=one>; rel="next"',
        },
        json=[{"id": "one"}],
        status_code=200,
    )
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/training-data?last_id=one",
        json=[{"id": "two"}],
        status_code=200,
    )

    td_list = simai_client.training_data.list()
    assert len(td_list) == 2
    assert td_list[0].id == "one"
    assert td_list[1].id == "two"


def test_training_data_list_with_filters(simai_client, httpx_mock):
    expected_query = urlencode(
        [
            (
                "filter[]",
                json.dumps(
                    {"field": "name", "operator": "EQ", "value": "thingo"},
                    separators=(",", ":"),
                ),
            ),
            (
                "filter[]",
                json.dumps(
                    {"field": "size", "operator": "LT", "value": "10000"},
                    separators=(",", ":"),
                ),
            ),
        ]
    )
    httpx_mock.add_response(
        method="GET",
        url=f"https://test.test/training-data?{expected_query}",
        headers={
            "X-Pagination": json.dumps({"total_pages": 1}),
        },
        json=[{"id": "one"}],
        status_code=200,
    )

    td = simai_client.training_data.list(filters=[("name", "EQ", "thingo"), ("size", "LT", 10000)])
    assert len(td) == 1
    assert len(httpx_mock.get_requests()) == 1


def test_training_data_add_to_project(
    simai_client, training_data_factory, project_factory, httpx_mock
):
    td: TrainingData = training_data_factory(id="08080")
    project = project_factory(id="09090")
    httpx_mock.add_response(
        method="PUT",
        url=f"https://test.test/training-data/{td.id}/project/{project.id}/association",
        status_code=204,
    )
    td.add_to_project(project)


def test_training_data_remove_from_project(
    simai_client, training_data_factory, project_factory, httpx_mock
):
    td: TrainingData = training_data_factory(id="08080")
    project = project_factory(id="09090")
    httpx_mock.add_response(
        method="DELETE",
        url=f"https://test.test/training-data/{td.id}/project/{project.id}/association",
        status_code=204,
    )
    td.remove_from_project(project)


@pytest.mark.parametrize(
    "td_factory_args",
    [
        ({"id": "777", "name": "ICBM", "subset": SubsetEnum.TRAINING}),
        ({"id": "888", "name": "Duke Nukem", "subset": "Training"}),
        ({"id": "999", "name": "Roman"}),
        ({"id": "81", "name": "Diablo", "subset": None}),
    ],
)
def test_get_subset(training_data_factory, project_factory, td_factory_args, httpx_mock):
    project = project_factory(id="e45y", name="coolest_proj")
    td_subset = td_factory_args.get("subset")
    td_factory_args["project"] = project
    td: TrainingData = training_data_factory(**td_factory_args)

    httpx_mock.add_response(
        method="GET",
        url=f"https://test.test/projects/{project.id}/data/{td.id}/subset",
        status_code=200,
        json={"subset": td_subset},
    )
    assert td.get_subset(project=project) == td_subset


def test_get_subset_fails_enum_check(training_data_factory, project_factory, httpx_mock):
    project = project_factory(id="bon5ai", name="coolest_proj")
    td: TrainingData = training_data_factory(project=project, id="415")
    td_subset = "Trainidation"
    httpx_mock.add_response(
        method="GET",
        url=f"https://test.test/projects/{project.id}/data/{td.id}/subset",
        status_code=200,
        json={"subset": td_subset},
    )
    with pytest.raises(ValueError) as e:
        td.get_subset(project=project)
    assert str(e.value) == f"'{td_subset}' is not a valid SubsetEnum"


def test_assign_subset(training_data_factory, project_factory, httpx_mock):
    project = project_factory(id="n07e45y", name="bananarama")
    td: TrainingData = training_data_factory(project=project, subset=SubsetEnum.TRAINING)

    for _ in range(3):
        httpx_mock.add_response(
            method="PUT",
            url=f"https://test.test/projects/{project.id}/data/{td.id}/subset",
            status_code=200,
            json={"subset": SubsetEnum.TEST},
        )
    td.assign_subset(project=project, subset=SubsetEnum.TEST)
    td.assign_subset(project=project, subset="Test")
    td.assign_subset(project=project, subset=None)
    with pytest.raises(InvalidArguments) as ve:
        td.assign_subset(project=project, subset="Travalignorinestidation")
    assert str(ve.value) == "Must be None or one of: 'Training', 'Test'."


def test_get_training_data_by_id(simai_client, httpx_mock):
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/training-data/td-123",
        json={"id": "td-123", "name": "My Training Data"},
        status_code=200,
    )
    td = simai_client.training_data.get(id="td-123")
    assert td.id == "td-123"
    assert td.name == "My Training Data"


def test_get_training_data_by_name(simai_client, httpx_mock):
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/training-data/name/My%20Training%20Data",
        json={"id": "td-456", "name": "My Training Data"},
        status_code=200,
    )
    td = simai_client.training_data.get(name="My Training Data")
    assert td.id == "td-456"
    assert td.name == "My Training Data"


def test_get_training_data_invalid_arguments(simai_client):
    with pytest.raises(InvalidArguments) as e:
        simai_client.training_data.get()
    assert str(e.value) == "Either 'id' or 'name' argument must be specified."

    with pytest.raises(InvalidArguments) as e:
        simai_client.training_data.get(id="td-123", name="My Training Data")
    assert str(e.value) == "Cannot specify both 'id' and 'name' arguments."


def test_rename_training_data(simai_client, httpx_mock, training_data_factory):
    td: TrainingData = training_data_factory(id="39646")
    httpx_mock.add_response(
        method="PATCH",
        url="https://test.test/training-data/39646",
        status_code=204,
    )
    # because rename triggers a reload of the object from the server
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/training-data/39646",
        json={"id": "39646", "name": "the new name"},
        status_code=200,
    )

    td.rename("the new name")

    assert td.id == "39646"
    assert td.name == "the new name"



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import pytest

from ansys.simai.core.data.model_configuration import (
    GlobalCoefficientDefinition,
    ModelConfiguration,
    ModelOutput,
    SupportedBuildPresets,
)
from ansys.simai.core.errors import InvalidArguments, ProcessingError

METADATA_RAW = {
    "boundary_conditions": {
        "fields": [
            {
                "format": "value",
                "keys": None,
                "name": "Vx",
                "unit": None,
                "value": -5.569587230682373,
            }
        ]
    },
    "surface": {
        "fields": [
            {
                "format": "value",
                "keys": None,
                "location": "cell",
                "name": "Pressure",
                "unit": None,
            },
            {
                "format": "value",
                "keys": None,
                "location": "cell",
                "name": "TurbulentViscosity",
                "unit": None,
            },
            {
                "format": "value",
                "keys": None,
                "location": "cell",
                "name": "WallShearStress_0",
                "unit": None,
            },
        ]
    },
    "volume": {
        "fields": [
            {
                "format": "value",
                "keys": None,
                "location": "cell",
                "name": "Pressure",
                "unit": None,
            },
            {
                "format": "value",
                "keys": None,
                "location": "cell",
                "name": "Velocity_0",
                "unit": None,
            },
            {
                "format": "value",
                "keys": None,
                "location": "cell",
                "name": "Velocity_1",
                "unit": None,
            },
            {
                "format": "value",
                "keys": None,
                "location": "cell",
                "name": "VolumeFractionWater",
                "unit": None,
            },
        ]
    },
}

SAMPLE_RAW = {
    "extracted_metadata": METADATA_RAW,
    "id": "DarkKnight",
    "is_complete": True,
    "is_deletable": True,
    "is_in_a_project_being_trained": False,
    "is_sample_of_a_project": True,
    "luggage_version": "52.2.2",
}


def test_build_preset_error(simai_client):
    """WHEN build_preset gets a non-supported value
    THEN an InvalidArgument is raised."""

    raw_project = {
        "id": "xX007Xx",
        "name": "fifi",
    }

    project = simai_client._project_directory._model_from(raw_project)

    with pytest.raises(InvalidArguments) as excinfo:
        ModelConfiguration(
            project=project,
            build_preset="not_valid_value",
        )

        assert f"{list(SupportedBuildPresets)}" in excinfo.value


def test_model_input_not_none(simai_client):
    """WHEN ModelConfiguration.input gets a None value
    THEN an InvalidArgument is raised."""

    raw_project = {
        "id": "xX007Xx",
        "name": "fifi",
    }

    project = simai_client._project_directory._model_from(raw_project)

    bld_conf = ModelConfiguration(
        project=project,
    )

    with pytest.raises(InvalidArguments):
        bld_conf.input = None


def test_model_output_not_none(simai_client):
    """WHEN ModelConfiguration.input gets a None value
    THEN an InvalidArgument is raised."""

    raw_project = {
        "id": "xX007Xx",
        "name": "fifi",
    }

    project = simai_client._project_directory._model_from(raw_project)

    bld_conf = ModelConfiguration(
        project=project,
    )

    with pytest.raises(InvalidArguments):
        bld_conf.output = None


def test_model_gc_same_name_as_output_scalar(simai_client):
    """WHEN global coefficient has the same name as a output scalar
    THEN an ProcessingError is raised."""

    raw_project = {
        "id": "xX007Xx",
        "name": "fifi",
    }

    project = simai_client._project_directory._model_from(raw_project)

    # with specific exception message
    with pytest.raises(ProcessingError) as e:
        ModelConfiguration(
            project=project,
            global_coefficients=[
                GlobalCoefficientDefinition("max(pressure)", "conflict_name", "cells")
            ],
            output=ModelOutput(scalars=["conflict_name"]),
        )
    assert (
        str(e.value)
        == "Global Coefficient 'conflict_name' has the same name as an output scalar variable."
    )


def test_model_output_scalars_same_name_gc(mocker, simai_client):
    """WHEN ModelConfiguration.output.scalars has the same name as a global coefficient
    THEN an ProcessingError is raised."""

    raw_project = {
        "id": "xX007Xx",
        "name": "fifi",
        "sample": SAMPLE_RAW,
    }

    project = simai_client._project_directory._model_from(raw_project)

    mocker.patch.object(project, "process_gc_formula", autospec=True)

    with pytest.raises(ProcessingError) as e:
        config = ModelConfiguration(
            project=project,
            global_coefficients=[
                GlobalCoefficientDefinition("max(pressure)", "conflict_name", "cells")
            ],
        )
        config.output = ModelOutput(scalars=["conflict_name"])
    assert (
        str(e.value)
        == "Scalar variables 'conflict_name' have the same name as global coefficients."
    )



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import TYPE_CHECKING, Literal, Optional, Union

from pydantic import BaseModel, Field, ValidationError, model_validator

from ansys.simai.core.data.base import ComputableDataModel, Directory
from ansys.simai.core.data.types import Identifiable, get_id_from_identifiable
from ansys.simai.core.errors import InvalidArguments

if TYPE_CHECKING:
    from ansys.simai.core.data.geomai.projects import GeomAIProject


class GeomAIModelConfiguration(BaseModel):
    nb_latent_param: int = Field(
        default=512,
        ge=2,
        le=1024,
    )
    """
    An **optional** integer that defines the dimension of the latent space between 2 and 1024.

    .. warning::

        The default value of 512 is highly recommended if you are a non-expert user.

    If you are an expert user, ready to experiment and assess correctly your models performance for different latent parameter value,
    and your use case has specific constraints, you can change this parameter, keeping the following in mind:

    - If the number is too low, geometries might become too coarse, and the reconstruction performance might degrade.
    - If the number is too high, the model will not be able to generate new geometries correctly.

    The default value of ``512`` is used if ``nb_latent_param`` is not set.
    """
    build_preset: Optional[Literal["debug", "short", "default", "long"]] = None
    """
    The preset to use for the model training duration. One of `debug`, `short`, `default`, `long`.

    - ``debug``: 4 minutes + 15 sec per geometry.
    - ``short``: 45 minutes + 15 sec per geometry.
    - ``default``: 3 hours + 15 sec per geometry.
    - ``long``: 15 hours + 15 sec per geometry.

    Mutually exclusive with ``nb_epochs``.
    """
    nb_epochs: Optional[int] = Field(default=None, ge=1, le=1000)
    """
    The number of times each training data is seen by the model during the training, between 1 and 1000.

    Mutually exclusive with ``build_preset``.
    """

    def __init__(self, *args, **kwargs):
        """Raises :exc:`~ansys.simai.core.errors.InvalidArguments` if the input data cannot be validated to from a valid model."""
        try:
            super().__init__(*args, **kwargs)
        except ValidationError as e:
            raise InvalidArguments(e.errors(include_url=False)) from None

    @model_validator(mode="after")
    def _build_preset_or_nb_epochs(self) -> "GeomAIModelConfiguration":
        if (self.nb_epochs is None and self.build_preset is None) or (
            self.nb_epochs is not None and self.build_preset is not None
        ):
            raise ValueError("Exactly one of nb_epochs or build_preset must have a value")
        return self


class GeomAIModel(ComputableDataModel):
    """GeomAI model representation."""

    def __repr__(self) -> str:
        return f"<Model: {self.id}>"

    @property
    def project_id(self) -> str:
        """The ID of the GeomAI project where the model exists."""
        return self.fields["project_id"]

    @property
    def configuration(self) -> GeomAIModelConfiguration:
        """Build configuration of a model."""
        return GeomAIModelConfiguration.model_construct(
            **self.fields["configuration"],
        )


class GeomAIModelDirectory(Directory[GeomAIModel]):
    """Provides a collection of methods related to building models."""

    _data_model = GeomAIModel

    def get(self, id: str) -> GeomAIModel:
        """[Do not use] Get a GeomAI model by ID.

        Args:
            id: ID of the model.
        """

        raise NotImplementedError(
            "The method 'get' of the class GeomAIModel is not implemented yet."
        )

    def build(
        self,
        project: Identifiable["GeomAIProject"],
        configuration: Union[dict, GeomAIModelConfiguration],
    ) -> GeomAIModel:
        """Launches a GeomAI build with the given configuration.

        Args:
            project: The GeomAI project in which to run the training.
            configuration: a :class:`GeomAIModelConfiguration` object that contains the properties to be used in the build.

        Examples:
            .. code-block:: python

                import ansys.simai.core as asc
                from ansys.simai.core.data.geomai.models import GeomAIModelConfiguration

                simai_client = asc.from_config()
                project = simai_client.geomai.projects.get("new_secret_project")
                configuration = GeomAIModelConfiguration(build_preset="default")
                model = simai_client.geomai.models.build(project, configuration)


            Use a previous configuration for a new build in the same project:

            .. code-block:: python

                a_project = simai.geomai.projects.get("project_A")
                build_conf = a_project.last_model_configuration
                new_model = simai.geomai.models.build(build_conf)

            Use a previous configuration for a new build in another project:

            .. code-block:: python

                a_project = simai.geomai.projects.get("project_A")
                build_conf = a_project.last_model_configuration
                b_project = simai.geomai.projects.get("project_B")
                new_model = simai.geomai.models.build(build_conf)

        """
        project_id = get_id_from_identifiable(project)
        configuration = (
            configuration
            if isinstance(configuration, GeomAIModelConfiguration)
            else GeomAIModelConfiguration(**configuration)
        )
        response = self._client._api.launch_geomai_build(
            project_id,
            configuration.model_dump(),
        )
        return self._model_from(response)



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json

import pytest

import ansys.simai.core
from ansys.simai.core.data import post_processings
from ansys.simai.core.data.predictions import Prediction
from ansys.simai.core.errors import InvalidArguments


def test_post_processing_prediction_attribute(post_processing_factory, httpx_mock):
    """WHEN accessing the prediction attribute of a PostProcessing
    THEN the prediction is returned
    """
    post_processing = post_processing_factory(type="GlobalCoefficients", prediction_id="java")
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/predictions/java",
        json={"id": "java", "state": "processing"},
        status_code=200,
    )

    assert isinstance(post_processing.prediction, Prediction)
    assert post_processing.prediction.id == "java"


def test_post_processing_call_prediction_attribute_twice(post_processing_factory, httpx_mock):
    """WHEN accessing the prediction attribute of a PostProcessing twice
    THEN the endpoint is called only once
    """
    post_processing = post_processing_factory(type="GlobalCoefficients", prediction_id="sumatra")
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/predictions/sumatra",
        json={"id": "sumatra", "state": "successful"},
        status_code=200,
    )

    post_processing.prediction  # noqa: B018
    post_processing.prediction  # noqa: B018

    assert len(httpx_mock.get_requests()) == 1


def test_post_processing_call_prediction_attribute_already_registered(
    prediction_factory, post_processing_factory, httpx_mock
):
    """WHEN accessing the prediction attribute of a PostProcessing when the geometry exists locally
    THEN no query is ran
    """
    prediction_factory(id="registered_prediction")
    post_processing = post_processing_factory(
        type="GlobalCoefficients", prediction_id="registered_prediction"
    )

    post_processing.prediction  # noqa: B018

    assert len(httpx_mock.get_requests()) == 0


def test_post_processing_ran_from_prediction_already_has_a_prediction(
    prediction_factory, post_processing_factory, httpx_mock
):
    """WHEN Running a post-treatment on a prediction
    THEN the prediction attribute is available without calling the /predictions/id endpoint
    """
    pred = prediction_factory(post_processings=[post_processing_factory(type="GlobalCoefficients")])
    global_coefficients = pred.post.global_coefficients()

    assert global_coefficients.prediction == pred
    assert len(httpx_mock.get_requests()) == 0


def test_post_processing_dont_run_exists_locally(
    prediction_factory, post_processing_factory, httpx_mock
):
    """WHEN Running a post-processing with run=False and it already exists locally
    THEN The local post-processing is returned and the API not called
    """
    pred = prediction_factory(post_processings=[post_processing_factory(type="GlobalCoefficients")])

    assert pred.post.global_coefficients(run=False) is not None

    assert len(httpx_mock.get_requests()) == 0


def test_post_processing_dont_run_does_not_exist_locally_no_params(prediction_factory, httpx_mock):
    """WHEN Running a parameter-less post-processing with run=False and it does not exist locally
    THEN The local post-processing is returned and the API get endpoint is called
    """
    pred = prediction_factory()
    httpx_mock.add_response(
        method="GET",
        url=f"https://test.test/predictions/{pred.id}/post-processings/GlobalCoefficients",
        json=[{"id": "0001", "type": "GlobalCoefficients"}],
        status_code=200,
    )

    pp = pred.post.global_coefficients(run=False)
    assert pp.id == "0001"


def test_post_processing_dont_run_does_not_exist_locally_with_params(
    prediction_factory, httpx_mock
):
    """WHEN Running a post-processing with parameters and run=False and it does not exist locally
    THEN The local post-processing is returned and the API get endpoint is called
    """
    pred = prediction_factory()
    httpx_mock.add_response(
        method="GET",
        url=f"https://test.test/predictions/{pred.id}/post-processings/SurfaceEvol?filters=%7B%22axis%22%3A%20%22x%22%2C%20%22delta%22%3A%200.5%7D",
        json=[
            {
                "id": "0002",
                "type": "SurfaceEvol",
                "location": {"axis": "x", "delta": 0.5},
            }
        ],
        status_code=200,
    )

    pp = pred.post.surface_evolution(axis="x", delta=0.5, run=False)
    assert pp.id == "0002"


def test_post_processing_dont_run_does_not_exist_locally_or_remotely(
    prediction_factory, httpx_mock
):
    """WHEN Running a post-processing with run=False and it does not exist locally or on the server
    THEN None is returned and the API get endpoint is called
    """
    pred = prediction_factory()
    httpx_mock.add_response(
        method="GET",
        url=f"https://test.test/predictions/{pred.id}/post-processings/GlobalCoefficients",
        json=[],
        status_code=200,
    )

    assert pred.post.global_coefficients(run=False) is None


def test_post_processing_run_from_directory(simai_client, prediction_factory, httpx_mock):
    """WHEN I run a post-processing from the directory
    THEN The created post-processing is returned
    """
    pred = prediction_factory()
    for _ in range(2):
        httpx_mock.add_response(
            method="GET",
            url=f"https://test.test/predictions/{pred.id}",
            json={"id": pred.id},
            status_code=200,
        )
    httpx_mock.add_response(
        method="POST",
        url=f"https://test.test/predictions/{pred.id}/post-processings/GlobalCoefficients",
        json={"id": "7777", "state": "successful"},
        status_code=200,
    )

    pp = simai_client.post_processings.run(ansys.simai.core.GlobalCoefficients, pred.id)
    pp_bis = simai_client.post_processings.run("GlobalCoefficients", pred.id)
    assert pp.id == "7777"
    assert pp_bis.id == "7777"


def test_post_processing_list(simai_client, httpx_mock):
    """WHEN simai.post_processings.list is called
    THEN it returns a list of all post-processings for the workspace
    """
    httpx_mock.add_response(
        method="GET",
        url=f"https://test.test/post-processings/?workspace={simai_client.current_workspace.id}",
        headers={"X-Pagination": json.dumps({"total_pages": 3})},
        json=[{"type": "GlobalCoefficients", "id": "1"}],
    )
    httpx_mock.add_response(
        method="GET",
        url=f"https://test.test/post-processings/?workspace={simai_client.current_workspace.id}&page=2",
        json=[{"type": "SurfaceVTP", "id": "2"}],
    )
    httpx_mock.add_response(
        method="GET",
        url=f"https://test.test/post-processings/?workspace={simai_client.current_workspace.id}&page=3",
        json=[{"type": "Slice", "id": "3"}],
    )

    pps = simai_client.post_processings.list()
    assert len(pps) == 3
    assert [pp.id for pp in pps] == ["1", "2", "3"]


def test_post_processing_list_prediction_and_workspace_forbidden(simai_client, httpx_mock):
    """WHEN simai.post_processings.list is called with a prediction and a workspace
    THEN InvalidArguments error is raised
    """
    with pytest.raises(InvalidArguments):
        simai_client.post_processings.list(workspace="toto", prediction="tata")


def test_post_processing_list_in_prediction(simai_client, mocker, httpx_mock):
    api_mock = mocker.Mock(return_value=[])
    simai_client._api.get_post_processings_for_prediction = api_mock
    pp = simai_client.post_processings.list(prediction="CrouAnthem")
    assert pp == []
    api_mock.assert_called_with("CrouAnthem", None)


def test_post_processing_list_in_workspace(simai_client, mocker, httpx_mock):
    api_mock = mocker.Mock(return_value=[])
    simai_client._api.get_post_processings_in_workspace = api_mock
    pp = simai_client.post_processings.list(
        workspace="BODEGA", post_processing_type=post_processings.Slice
    )
    assert pp == []
    api_mock.assert_called_with("BODEGA", post_processings.Slice._api_name())


def test_prediction_post_list(simai_client, mocker, prediction_factory, httpx_mock):
    pred = prediction_factory()
    api_mock = mocker.Mock(return_value=[])
    simai_client._api.get_post_processings_for_prediction = api_mock
    pp = pred.post.list(post_processings.SurfaceVTP)
    assert pp == []
    api_mock.assert_called_with(pred.id, post_processings.SurfaceVTP._api_name())



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from json import JSONDecodeError

import httpx
import pytest

from ansys.simai.core.errors import ApiClientError, SimAIError
from ansys.simai.core.utils.requests import handle_http_errors, handle_response


def test_handle_http_errors(mocker):
    response_json_mock = mocker.patch("httpx.Response.json")
    raise_for_status_mock = mocker.patch("httpx.Response.raise_for_status")

    # Error without json body
    raise_for_status_mock.side_effect = httpx.HTTPError("woaw")
    response_json_mock.side_effect = ValueError()

    with pytest.raises(SimAIError):
        handle_http_errors(httpx.Response(200))

    # Error with json body
    raise_for_status_mock.side_effect = httpx.HTTPError("boadyful")
    response_json_mock.side_effect = None
    response_json_mock.return_value = {"status": "rekt"}

    with pytest.raises(SimAIError, match="rekt"):
        handle_http_errors(httpx.Response(200))


@pytest.mark.parametrize(
    "status_code, return_value, return_json",
    (
        (200, {"status": "succeeding"}, True),
        (200, {"status": "succeeding"}, False),
        (204, "Should be None", True),
    ),
)
def test_handle_response_success(mocker, status_code, return_value, return_json):
    mocker.patch("ansys.simai.core.utils.requests.handle_http_errors")
    mocker.patch.object(httpx.Response, "json", return_value=return_value)

    result = handle_response(httpx.Response(status_code), return_json=return_json)

    if return_json is False:
        assert isinstance(result, httpx.Response)
    elif status_code == 204:
        assert result is None
    else:
        assert result == return_value


@pytest.mark.parametrize(
    "status_code, side_effect",
    (
        (200, JSONDecodeError("msg", "doc", 0)),
        (200, ValueError()),
    ),
)
def test_handle_response_raises(mocker, status_code, side_effect):
    mocker.patch("ansys.simai.core.utils.requests.handle_http_errors")
    mocker.patch("httpx.Response.json", side_effect=side_effect)

    with pytest.raises(ApiClientError):
        handle_response(httpx.Response(status_code), return_json=True)



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest

from ansys.simai.core.data.types import Range


@pytest.fixture
def geometries_with_two_metadata():
    """A list of geometries with 2 numerical metadata:
    - length: 10 values ranging from 0.0 to 90.0
    - width: 10 values ranging from 0 to 9
    """

    def query_geometries_data(workspace_id, server_filters=None):
        geometries = []
        for i in range(10):
            for j in range(10):
                geometries.append(
                    {"id": f"l-{i}-w-{j}", "metadata": {"length": i * 10.0, "width": j}}
                )
        return geometries

    yield query_geometries_data


@pytest.fixture(scope="function")
def extra_client(simai_client, geometries_with_two_metadata):
    simai_client._api.geometries = geometries_with_two_metadata
    yield simai_client


def test_no_filter(extra_client):
    """WHEN getting geometries with no Range filter
    THEN all geometries are returned
    """
    geometries = extra_client.geometries.list()
    assert len(geometries) == 100

    geometries = extra_client.geometries.list(filters={})
    assert len(geometries) == 100


def test_filtering_range_simple(extra_client):
    """WHEN getting geometries with filtering with a Range
    THEN geometries with metadata >= min and < max are returned
    """
    geometries = extra_client.geometries.list(filters={"length": Range(27, 45)})
    assert len(geometries) == 20
    for geometry in geometries:
        assert geometry.metadata["length"] in [30, 40]


def test_filtering_range_bounds_inclusion(extra_client):
    """WHEN filtering geometries with a Range
    THEN both bounds are included in results
    """
    geometries = extra_client.geometries.list(filters={"length": Range(30, 40)})
    assert len(geometries) == 20
    for geometry in geometries:
        assert geometry.metadata["length"] in [30, 40]


def test_filtering_same_min_max(extra_client):
    """WHEN filtering with a Range whose bounds are equal
    THEN only geometries matching bounds are returned
    """
    geometries = extra_client.geometries.list(filters={"length": Range(20, 20)})
    assert len(geometries) == 10

    geometries = extra_client.geometries.list(filters={"length": Range(46, 46)})
    assert len(geometries) == 0


def test_filtering_two_metadata(extra_client):
    """WHEN filtering with Range objects on two metadata
    THEN geometries whose metadata each matching its Range are returned
    """
    geometries = extra_client.geometries.list(
        filters={"length": Range(35, 55), "width": Range(7.0, 9.0)}
    )
    assert len(geometries) == 6
    for geometry in geometries:
        assert geometry.metadata["length"] in [40, 50]
        assert geometry.metadata["width"] in [7, 8, 9]


def test_filtering_with_only_lower_bound(extra_client):
    """WHEN filtering with a Range having only a lower bound
    THEN geometries whose metadata >= this bound are returned
    """
    geometries = extra_client.geometries.list(filters={"length": Range(min=80)})
    assert len(geometries) == 20
    for geometry in geometries:
        assert geometry.metadata["length"] in [80, 90]


def test_filtering_with_only_upper_bound(extra_client):
    """WHEN filtering with a Range having only a upper bound
    THEN geometries whose metadata <= this bound are returned
    """
    geometries = extra_client.geometries.list(filters={"length": Range(max=40)})
    assert len(geometries) == 50
    for geometry in geometries:
        assert geometry.metadata["length"] in [0, 10, 20, 30, 40]


def test_filtering_tolerance_upper(extra_client):
    """WHEN filtering with Ranges with a tolerance
    THEN values just above the min bound are included
    AND values just above to the max bound are included
    """
    epsilon = 5 * 10**-2
    geometries = extra_client.geometries.list(
        filters={"length": Range(30 + epsilon, 50 + epsilon, tolerance=0.1)}
    )
    assert len(geometries) == 30
    for geometry in geometries:
        assert geometry.metadata["length"] in [30, 40, 50]


def test_filtering_default_tolerance_upper(extra_client):
    """WHEN filtering with Ranges without passing a tolerance
    THEN tolerance of 10**-6 is used
    AND values just above the min bound are included
    AND values just above to the max bound are included
    """
    epsilon = 5 * 10**-7
    geometries = extra_client.geometries.list(filters={"length": Range(30 + epsilon, 50 + epsilon)})
    assert len(geometries) == 30
    for geometry in geometries:
        assert geometry.metadata["length"] in [30, 40, 50]


def test_filtering_above_tolerance(extra_client):
    """WHEN filtering with Ranges using default 10**-6 tolerance
    THEN values differing more than tolerance are considered different
    """
    epsilon = 5 * 10**-5
    geometries = extra_client.geometries.list(filters={"length": Range(30 + epsilon, 50 + epsilon)})
    assert len(geometries) == 20
    for geometry in geometries:
        assert geometry.metadata["length"] in [40, 50]


def test_filtering_tolerance_lower(extra_client):
    """WHEN filtering with Ranges with a given tolerance
    THEN values just below the min bound are included
    AND values just below the max bound are included
    """
    epsilon = 5 * 10**-2
    geometries = extra_client.geometries.list(
        filters={"length": Range(30 - epsilon, 50 - epsilon, tolerance=0.1)}
    )
    assert len(geometries) == 30
    for geometry in geometries:
        assert geometry.metadata["length"] in [30, 40, 50]


def test_filtering_default_tolerance_lower(extra_client):
    """WHEN filtering with Ranges without passing a tolerance
    THEN tolerance of 10**-6 is used
    AND values just below the min bound are included
    AND values just below the max bound are included
    """
    epsilon = 5 * 10**-7
    geometries = extra_client.geometries.list(filters={"length": Range(30 - epsilon, 50 - epsilon)})
    assert len(geometries) == 30
    for geometry in geometries:
        assert geometry.metadata["length"] in [30, 40, 50]



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import math

import pytest

from ansys.simai.core.errors import SimAIError
from ansys.simai.core.utils.numerical import (
    cast_values_to_float,
    convert_axis_and_coordinate_to_plane_eq_coeffs,
    is_bigger_or_equal_with_tolerance,
    is_bigger_with_tolerance,
    is_equal_with_tolerance,
    is_number,
    is_smaller_or_equal_with_tolerance,
    is_smaller_with_tolerance,
)


def test_is_number():
    assert is_number(0)
    assert is_number(3)
    assert is_number(-3)
    assert is_number(10**8)
    assert is_number(10**-7)
    assert not is_number(math.nan)
    assert not is_number(None)
    assert not is_number("10")
    assert not is_number([10])
    assert not is_number({10})
    assert not is_number({})


def test_is_equal_with_tolerance():
    assert not is_equal_with_tolerance(1, 3)
    assert not is_equal_with_tolerance(3, 1)
    assert not is_equal_with_tolerance(10**-4, 10**-3)
    assert not is_equal_with_tolerance(10**-3, 10**-4)
    assert not is_equal_with_tolerance(-3, -1)
    assert not is_equal_with_tolerance(-1, -3)
    assert is_equal_with_tolerance(0, 0)
    assert is_equal_with_tolerance(111, 111)

    # Test tolerance
    a = 55
    assert is_equal_with_tolerance(a - 10**-7, a)
    assert not is_equal_with_tolerance(a - 10**-5, a)
    assert is_equal_with_tolerance(a - 10**-5, a, tolerance=10**-4)

    assert is_equal_with_tolerance(a + 10**-7, a)
    assert not is_equal_with_tolerance(a + 10**-5, a)
    assert is_equal_with_tolerance(a + 10**-5, a, tolerance=10**-4)


def test_is_smaller_with_tolerance():
    assert is_smaller_with_tolerance(1, 3)
    assert not is_smaller_with_tolerance(3, 1)
    assert is_smaller_with_tolerance(10**-4, 10**-3)
    assert not is_smaller_with_tolerance(10**-3, 10**-4)
    assert is_smaller_with_tolerance(-3, -1)
    assert not is_smaller_with_tolerance(-1, -3)
    assert not is_smaller_with_tolerance(0, 0)
    assert not is_smaller_with_tolerance(111, 111)

    # Test tolerance
    a = 55
    assert not is_smaller_with_tolerance(a - 10**-7, a)
    assert is_smaller_with_tolerance(a - 10**-5, a)
    assert not is_smaller_with_tolerance(a - 10**-5, a, tolerance=10**-4)

    assert not is_smaller_with_tolerance(a + 10**-7, a)
    assert not is_smaller_with_tolerance(a + 10**-5, a)
    assert not is_smaller_with_tolerance(a + 10**-5, a, tolerance=10**-4)


def test_is_smaller_or_equal_with_tolerance():
    assert is_smaller_or_equal_with_tolerance(1, 3)
    assert not is_smaller_or_equal_with_tolerance(3, 1)
    assert is_smaller_or_equal_with_tolerance(10**-4, 10**-3)
    assert not is_smaller_or_equal_with_tolerance(10**-3, 10**-4)
    assert is_smaller_or_equal_with_tolerance(-3, -1)
    assert not is_smaller_or_equal_with_tolerance(-1, -3)
    assert is_smaller_or_equal_with_tolerance(0, 0)
    assert is_smaller_or_equal_with_tolerance(111, 111)

    # Test tolerance
    a = 55
    assert is_smaller_or_equal_with_tolerance(a - 10**-7, a)
    assert is_smaller_or_equal_with_tolerance(a - 10**-5, a)
    assert is_smaller_or_equal_with_tolerance(a - 10**-5, a, tolerance=10**-4)

    assert is_smaller_or_equal_with_tolerance(a + 10**-7, a)
    assert not is_smaller_or_equal_with_tolerance(a + 10**-5, a)
    assert is_smaller_or_equal_with_tolerance(a + 10**-5, a, tolerance=10**-4)


def test_is_bigger_with_tolerance():
    assert not is_bigger_with_tolerance(1, 3)
    assert is_bigger_with_tolerance(3, 1)
    assert not is_bigger_with_tolerance(10**-4, 10**-3)
    assert is_bigger_with_tolerance(10**-3, 10**-4)
    assert not is_bigger_with_tolerance(-3, -1)
    assert is_bigger_with_tolerance(-1, -3)
    assert not is_bigger_with_tolerance(0, 0)
    assert not is_bigger_with_tolerance(111, 111)

    # Test tolerance
    a = 55
    assert not is_bigger_with_tolerance(a - 10**-7, a)
    assert not is_bigger_with_tolerance(a - 10**-5, a)
    assert not is_bigger_with_tolerance(a - 10**-5, a, tolerance=10**-4)

    assert not is_bigger_with_tolerance(a + 10**-7, a)
    assert is_bigger_with_tolerance(a + 10**-5, a)
    assert not is_bigger_with_tolerance(a + 10**-5, a, tolerance=10**-4)


def test_is_bigger_or_equal_with_tolerance():
    assert not is_bigger_or_equal_with_tolerance(1, 3)
    assert is_bigger_or_equal_with_tolerance(3, 1)
    assert not is_bigger_or_equal_with_tolerance(10**-4, 10**-3)
    assert is_bigger_or_equal_with_tolerance(10**-3, 10**-4)
    assert not is_bigger_or_equal_with_tolerance(-3, -1)
    assert is_bigger_or_equal_with_tolerance(-1, -3)
    assert is_bigger_or_equal_with_tolerance(0, 0)
    assert is_bigger_or_equal_with_tolerance(111, 111)

    # Test tolerance
    a = 55
    assert is_bigger_or_equal_with_tolerance(a - 10**-7, a)
    assert not is_bigger_or_equal_with_tolerance(a - 10**-5, a)
    assert is_bigger_or_equal_with_tolerance(a - 10**-5, a, tolerance=10**-4)

    assert is_bigger_or_equal_with_tolerance(a + 10**-7, a)
    assert is_bigger_or_equal_with_tolerance(a + 10**-5, a)
    assert is_bigger_or_equal_with_tolerance(a + 10**-5, a, tolerance=10**-4)


def test_convert_axis_and_coordinate_to_plane_eq_coeffs():
    assert convert_axis_and_coordinate_to_plane_eq_coeffs("x", 10) == (1, 0, 0, 10)
    assert convert_axis_and_coordinate_to_plane_eq_coeffs("y", 10) == (0, 1, 0, 10)
    assert convert_axis_and_coordinate_to_plane_eq_coeffs("z", 10) == (0, 0, 1, 10)
    assert convert_axis_and_coordinate_to_plane_eq_coeffs("z", 100.5) == (
        0,
        0,
        1,
        100.5,
    )


def test_cast_values_to_float():
    input = {"convert": "NaN", "also_convert": "-inf"}
    output = cast_values_to_float(input)
    assert math.isnan(output["convert"])
    assert math.isinf(output["also_convert"])


def test_cast_values_to_float_invalid_type_raises():
    input = {"convert": "woops"}
    with pytest.raises(SimAIError):
        cast_values_to_float(input)



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import itertools

from ansys.simai.core.utils.grouping import _ToleranceGroup, _ToleranceGrouper
from ansys.simai.core.utils.numerical import is_equal_with_tolerance


def test_tolerance_grouper_basic():
    """WHEN using _ToleranceGrouper with a basic list and default or negligible tolerance,
    THEN each group contain distinct elements in the data list
    ALSO, forced_central_value has no effect.
    """
    data = [1, 1, 2, 3, 3, 3, 4, 5]
    for tolerance in [None, 10**-7, 10**-1]:
        for forced_central_value in [None, 2, 2.1, 22]:
            grouper = _ToleranceGrouper(
                forced_central_value=forced_central_value, tolerance=tolerance
            )
            result_groups = itertools.groupby(data, grouper)

            for i, group in enumerate(result_groups):
                num = i + 1
                bucket, items = group
                assert isinstance(bucket, _ToleranceGroup)
                assert bucket.center == num

                if num == 1:
                    assert list(items) == 2 * [num]
                elif num == 3:
                    assert list(items) == 3 * [num]
                else:
                    assert list(items) == [num]


def test_tolerance_grouper_tolerance():
    """WHEN using tolerance grouper with tolerance
    THEN groups of values that can all be grouped around a central value
    (approx. equal to this central value) are grouped together
    """
    data = [1, 1.001, 1.996, 2, 2.005, 4, 4, 4.005, 4.011, 4.019, 4.996, 5.002]
    tolerance = 10**-2
    grouper = _ToleranceGrouper(tolerance=tolerance)
    result_groups = itertools.groupby(data, grouper)

    for i, group in enumerate(result_groups):
        bucket, items = group
        assert isinstance(bucket, _ToleranceGroup)
        items = list(items)

        if i == 0:
            assert items == [1, 1.001]
            assert is_equal_with_tolerance(bucket.center, 1, tolerance=tolerance)
        elif i == 1:
            assert items == [1.996, 2, 2.005]
            assert is_equal_with_tolerance(bucket.center, 2, tolerance=tolerance)
        elif i == 2:
            # This bucket contains all those items,
            # as we can create a bucket around 4.005 : 4 ~= 4.005 ~= 4.011
            assert items == [4, 4, 4.005, 4.011]
            assert bucket.center == 4.005
        elif i == 3:
            assert items == [4.019]
            assert bucket.center == 4.019
        else:
            assert items == [4.996, 5.002]
            assert is_equal_with_tolerance(bucket.center, 5, tolerance=tolerance)


def test_tolerance_grouper_tolerance_with_forced_central_value():
    """WHEN using tolerance grouper with forced central value
    THEN those forced value must not be further than epsilon
        from their own bucket boundaries
    """
    data = [4, 4, 4.005, 4.011, 4.019]
    tolerance = 10**-2
    for forced_central_value in [4, 4.011]:
        grouper = _ToleranceGrouper(tolerance=tolerance, forced_central_value=forced_central_value)
        result_groups = itertools.groupby(data, grouper)

        for i, group in enumerate(result_groups):
            bucket, items = group
            assert isinstance(bucket, _ToleranceGroup)

            if i == 0:
                assert list(items) == [4, 4, 4.005]
                if forced_central_value == 4:
                    assert bucket.center == 4
            else:
                assert list(items) == [4.011, 4.019]



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
from typing import Any, BinaryIO, List, Optional, Tuple, Union

from pydantic import BaseModel, PositiveInt, ValidationError

from ansys.simai.core.data.base import ComputableDataModel, Directory
from ansys.simai.core.data.geomai.workspaces import GeomAIWorkspace
from ansys.simai.core.data.types import (
    File,
    Identifiable,
    get_id_from_identifiable,
)
from ansys.simai.core.errors import InvalidArguments

logger = logging.getLogger(__name__)


class GeomAIPredictionConfiguration(BaseModel):
    """The configuration used to run a GeomAI prediction."""

    latent_params: List[float]
    """A list of floats that represent the position of the geometry in the latent space.

    These parameters describe the shape in a compressed form.
    The number of floats should match the ``nb_latent_param`` your model was requested with.

    Required.
    """
    resolution: Optional[Tuple[PositiveInt, PositiveInt, PositiveInt]] = None
    """A list of three integers defining the number of voxels along the X, Y, and Z axes.

    Use higher resolution for complex or precise geometries, and lower resolution for simple shapes or quick previews.

    The total number of voxels must not exceed 900^3, that is `x`, `y`, `z` multiplied together must be less than or equal to 900^3.
    If you exceed that value, an error will occur.

    Defaults to ``[100,100,100]``, if ``None`` is provided.

    For the maximum resolution of 900^3, the prediction takes approximately 10 minutes (approximately 1 microsecond per voxel).
    """

    def __init__(self, *args, **kwargs):
        """Raises :exc:`~ansys.simai.core.errors.InvalidArguments` if the input data cannot be validated to from a valid model."""
        try:
            super().__init__(*args, **kwargs)
        except ValidationError as e:
            raise InvalidArguments(e.errors(include_url=False)) from None


class GeomAIPrediction(ComputableDataModel):
    """Provides the local representation of a GeomAI prediction object."""

    @property
    def configuration(self) -> GeomAIPredictionConfiguration:
        """The configuration used to run the prediction."""
        return GeomAIPredictionConfiguration.model_construct(**self.fields["configuration"])

    def delete(self) -> None:
        """Remove a prediction from the server."""
        self._client._api.delete_geomai_prediction(self.id)
        self._unregister()

    def download(self, file: Optional[File] = None) -> Union[None, BinaryIO]:
        """Download the file generated by the prediction.

        The downloaded file will be in the VTP format.

        Args:
            file: Binary file-object or the path of the file to put the content into.

        Returns:
            ``None`` if a file is specified or a binary file-object otherwise.
        """
        self.wait()
        return self._client._api.download_geomai_prediction(self.id, file)


class GeomAIPredictionDirectory(Directory[GeomAIPrediction]):
    """Provides a collection of methods related to GeomAI model predictions.

    This method is accessed through ``client.geomai.predictions``.

    Example:
        .. code-block:: python

            import ansys.simai.core as asc

            simai_client = asc.from_config()
            simai_client.geomai.predictions.list()

    """

    _data_model = GeomAIPrediction

    def list(
        self, workspace: Optional[Identifiable[GeomAIWorkspace]] = None
    ) -> List[GeomAIPrediction]:
        """List all GeomAI predictions on the server that belong to the specified workspace or the configured one.

        Args:
            workspace: ID or :class:`model <.workspaces.GeomAIWorkspace>` of the workspace to list the predictions for.
                This parameter is necessary if no workspace is set for the client.
        """
        workspace_id = get_id_from_identifiable(
            workspace, default=self._client.geomai._current_workspace
        )
        raw_predictions = self._client._api.geomai_predictions(workspace_id)
        return list(map(self._model_from, raw_predictions))

    def get(self, id: str) -> GeomAIPrediction:
        """Get a specific GeomAI prediction object from the server by ID.

        Args:
            id: ID of the prediction.

        Returns:
            :class:`GeomAIPrediction` instance with the given ID if it exists.

        Raises:
            NotFoundError: No prediction with the given ID exists.
        """
        return self._model_from(self._client._api.get_geomai_prediction(id))

    def delete(self, prediction: Identifiable[GeomAIPrediction]) -> None:
        """Delete a specific GeomAI prediction from the server.

        Args:
            prediction: ID or :class:`model <GeomAIPrediction>` of the prediction.

        Raises:
            NotFoundError: No prediction with the given ID exists.
        """
        prediction_id = get_id_from_identifiable(prediction)
        self._client._api.delete_geomai_prediction(prediction_id)
        self._unregister_item_with_id(prediction_id)

    def run(  # noqa: D417
        self,
        configuration: Union[GeomAIPredictionConfiguration, dict[str, Any]],
        workspace: Optional[Identifiable[GeomAIWorkspace]] = None,
    ) -> GeomAIPrediction:
        """Run a prediction in the given workspace with the given configuration.

        Args:
            configuration: The configuration to run the prediction with.
            workspace: Optional ID or :class:`model <.workspaces.GeomAIWorkspace>` of the target workspace.
                Defaults to the current workspace if set.

        Returns:
            Created prediction object.

        Raises:
            ProcessingError: If the server failed to process the request.

        Examples:
            .. code-block:: python

                import ansys.simai.core as asc

                simai_client = asc.from_config()
                workspace = simai_client.geomai.workspaces.list()[0]
                prediction = simai_client.geomai.predictions.run(
                    dict(latent_params=[0.1, 1.2, 0.76], resolution=(100, 100, 100)),
                    workspace,
                )

        """
        if not isinstance(configuration, GeomAIPredictionConfiguration):
            try:
                configuration = GeomAIPredictionConfiguration.model_validate(configuration)
            except ValidationError as e:
                raise InvalidArguments(e.errors(include_url=False)) from None
        workspace_id = get_id_from_identifiable(
            workspace, default=self._client.geomai._current_workspace
        )
        return self._model_from(
            self._client._api.run_geomai_prediction(
                workspace_id, configuration.model_dump(exclude_unset=True)
            )
        )

    def download(
        self, prediction: Identifiable[GeomAIPrediction], file: Optional[File] = None
    ) -> Union[None, BinaryIO]:
        """Download the file generated by the prediction.

        The downloaded file will be in the VTP format.

        Args:
            prediction: ID or :class:`model <GeomAIPrediction>` of the prediction.
            file: Binary file-object or the path of the file to put the content into.

        Returns:
            ``None`` if a file is specified or a binary file-object otherwise.

        Note:
            Consider using :meth:`GeomAIPrediction.download` instead which will wait for the
                prediction to be ready instead of throwing an error if it's not.
        """
        prediction_id = get_id_from_identifiable(prediction)
        return self._client._api.download_geomai_prediction(prediction_id, file)



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from io import BytesIO

from ansys.simai.core.data.post_processings import (
    CustomVolumePointCloud,
    DownloadableResult,
    GlobalCoefficients,
    Slice,
    SurfaceEvolution,
    SurfaceVTP,
    SurfaceVTPTDLocation,
    VolumeVTU,
)


def test_post_processing_result_global_coefficients(simai_client, httpx_mock):
    """WHEN Running a GlobalCoefficients post-processing on a prediction and calling its .data field,
    THEN a GET request is made on the post-processings/<id> endpoint
    ALSO the .data attribute is a dictionary containing the GlobalCoefficients data
    ALSO multiple accesses to .data don't call the endpoint multiple times
    """
    # Mock request for PP creation
    httpx_mock.add_response(
        method="POST",
        url="https://test.test/predictions/7546/post-processings/GlobalCoefficients",
        json={"id": "7167", "state": "successful"},
        status_code=200,
    )
    # Mock request for PP data
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/post-processings/7167",
        json={
            "data": {
                "values": {
                    "IsoStaticPressureForce": {
                        "data": {
                            "X": 4.441544089820862,
                            "Y": 0.32894839148911376,
                            "Z": 14.442021446988662,
                        },
                        "type": "base",
                        "unit": {"N": 1},
                    },
                    "WallShearStress": {
                        "data": {
                            "X": -0.6887743615741067,
                            "Y": 0.08958696618025093,
                            "Z": 0.0741612935822976,
                        },
                        "type": "base",
                        "unit": {"N": 1},
                    },
                }
            },
            "id": "ak2gr7pw",
            "location": {},
            "prediction": "bw6z91zj",
            "type": "GlobalCoefficients",
        },
        status_code=200,
    )

    pred = simai_client._prediction_directory._model_from({"id": "7546", "state": "successful"})
    global_coefficients = pred.post.global_coefficients()

    assert isinstance(global_coefficients, GlobalCoefficients)
    assert isinstance(global_coefficients.data, dict)
    assert len(global_coefficients.data) == 2
    assert global_coefficients.data["IsoStaticPressureForce"]["data"]["X"] == 4.441544089820862
    assert global_coefficients.data["IsoStaticPressureForce"]["type"] == "base"

    # check despite 2 access to global_coefficients.data, only 1 call to API endpoint
    requests = httpx_mock.get_requests()
    assert len([r for r in requests if r.url == "https://test.test/post-processings/7167"]) == 1


def test_post_processing_result_surface_evolution(simai_client, httpx_mock):
    """WHEN Running a SurfaceEvolution post-processing on a prediction and calling its .data field,
    THEN a GET request is made on the post-processings/<id> endpoint
    ALSO the .data attribute is a dictionary containing the expected data
    ALSO multiple accesses to .data don't call the endpoint multiple times
    """
    # Mock request for PP creation
    httpx_mock.add_response(
        method="POST",
        url="https://test.test/predictions/r26g04j8/post-processings/SurfaceEvol",
        json={"id": "4c7r4c", "state": "successful"},
        status_code=200,
    )
    for _ in range(2):
        # Mock request for PP data
        httpx_mock.add_response(
            method="GET",
            url="https://test.test/post-processings/4c7r4c",
            json={
                "data": {"resources": {"json": "https://s3.test/some/path/to/json"}},
                "id": "xn49z2z5",
                "location": {"axis": "x", "delta": 5},
                "prediction": "r26g04j8",
                "type": "SurfaceEvol",
            },
            status_code=200,
        )
    httpx_mock.add_response(
        method="GET",
        url="https://s3.test/some/path/to/json",
        json={
            "area": {
                "data": [
                    247.03766026765066,
                    307.41345249496834,
                    266.0681832190391,
                    62.77419701451614,
                ],
                "type": "base",
                "unit": {"m": 2},
            },
            "position": {
                "data": [
                    -75.07807159423828,
                    -52.07807159423828,
                    -29.07807159423828,
                    -6.078071594238281,
                ],
                "type": "base",
                "unit": {"m": 1},
            },
            "WallShearStress": {
                "data": {
                    "X": [
                        107167.4091582841,
                        131679.1151319072,
                        112707.19948165954,
                        25602.98322615308,
                    ],
                    "Y": [
                        -5719.545388303992,
                        -5059.645424596727,
                        -3057.985328271491,
                        -479.55662520016926,
                    ],
                    "Z": [
                        -2765.7740285333916,
                        -3634.3034712755143,
                        -2185.0554359153575,
                        -537.9401310387967,
                    ],
                },
                "type": "base",
                "unit": {"N": 1, "m": -2},
            },
        },
        status_code=200,
    )

    pred = simai_client._prediction_directory._model_from({"id": "r26g04j8", "state": "successful"})
    surface_evolution = pred.post.surface_evolution(axis="x", delta=5)

    assert isinstance(surface_evolution, SurfaceEvolution)
    assert isinstance(surface_evolution.data, DownloadableResult)
    surface_evolution_dict = surface_evolution.as_dict()
    assert len(surface_evolution_dict) == 3
    assert surface_evolution_dict["WallShearStress"]["data"]["X"][0] == 107167.4091582841

    # we have used surface_evolution.data twice, generating 2 hits to the endpoint
    requests = httpx_mock.get_requests()
    assert len([r for r in requests if r.url == "https://test.test/post-processings/4c7r4c"]) == 2


def test_post_processing_result_slice(simai_client, httpx_mock):
    """WHEN Running a Slice post-processing on a prediction and calling its .data field,
    THEN a GET request is made on the post-processings/<id> endpoint
    ALSO the .data attribute is a DownloadableResult object,
        that we can call a .binary_io method on,
        which will send the expected content
    """
    location = {"plane": [0, 1, 0, 20], "output_format": "png"}
    # Mock request for PP creation
    httpx_mock.add_response(
        method="POST",
        url="https://test.test/predictions/7546/post-processings/Slice",
        json={"id": "98413534", "state": "successful", "location": location},
        status_code=200,
    )
    for _ in range(2):
        # Mock request for PP data
        httpx_mock.add_response(
            method="GET",
            url="https://test.test/post-processings/98413534",
            json={
                "data": {"resources": {"png": "https://s3.test.test/slice.png?key=9988776655"}},
                "id": "98413534",
                "location": location,
                "prediction": "7546",
                "type": "Slice",
            },
            status_code=200,
        )
    # Mock request for slice binary download
    httpx_mock.add_response(
        method="GET",
        url="https://s3.test.test/slice.png?key=9988776655",
        content=b"this-is-slice-binary-data",
        headers={"Content-Type": "application/octet-stream"},
        status_code=200,
    )

    pred = simai_client._prediction_directory._model_from({"id": "7546", "state": "successful"})
    slice = pred.post.slice(axis="y", coordinate="20")

    assert isinstance(slice, Slice)
    assert isinstance(slice.data, DownloadableResult)
    # Test a binary download
    binary_io = slice.data.in_memory()
    assert isinstance(binary_io, BytesIO)
    data_line = binary_io.readline()
    assert data_line.decode("ascii") == "this-is-slice-binary-data"

    # we have used slice.data twice, generating 2 hits to the endpoint
    requests = httpx_mock.get_requests()
    assert len([r for r in requests if r.url == "https://test.test/post-processings/98413534"]) == 2


def test_post_processing_result_volume_vtu(simai_client, httpx_mock):
    """WHEN Running a VolumeVTU post-processing on a prediction and calling its .data field,
    THEN a GET request is made on the post-processings/<id> endpoint
    ALSO the .data attribute is a DownloadableResult object
        we can call a .binary_io method on
    """
    # Mock request for PP creation
    httpx_mock.add_response(
        method="POST",
        url="https://test.test/predictions/7546/post-processings/VolumeVTU",
        json={"id": "01010101654", "state": "successful"},
        status_code=200,
    )
    for _ in range(2):
        # Mock request for PP data
        httpx_mock.add_response(
            method="GET",
            url="https://test.test/post-processings/01010101654",
            json={
                "data": {"resources": {"vtk": "https://s3.test.test/slice.npz?key=445461321"}},
                "id": "01010101654",
                "location": {},
                "prediction": "7546",
                "type": "VolumeVTU",
            },
            status_code=200,
        )
    # Mock request for npz binary download
    httpx_mock.add_response(
        method="GET",
        url="https://s3.test.test/slice.npz?key=445461321",
        content=b"this-is-vtu-binary-data",
        headers={"Content-Type": "application/octet-stream"},
        status_code=200,
    )

    pred = simai_client._prediction_directory._model_from({"id": "7546", "state": "successful"})
    volume_vtu = pred.post.volume_vtu()
    assert isinstance(volume_vtu, VolumeVTU)
    assert isinstance(volume_vtu.data, DownloadableResult)
    # Test a binary download
    binary_io = volume_vtu.data.in_memory()
    assert isinstance(binary_io, BytesIO)
    data_line = binary_io.readline()
    assert data_line.decode("ascii") == "this-is-vtu-binary-data"

    # we have used vtu.data twice, generating 2 hits to the endpoint
    requests = httpx_mock.get_requests()
    assert (
        len([r for r in requests if r.url == "https://test.test/post-processings/01010101654"]) == 2
    )


def test_post_processing_result_surface_vtp(simai_client, httpx_mock):
    """WHEN Running a SurfaceVTP post-processing on a prediction and calling its .data field,
    THEN a GET request is made on the post-processings/<id> endpoint
    ALSO the .data attribute is a DownloadableResult object
        we can call a .binary_io method on
    """
    # Mock request for PP creation
    httpx_mock.add_response(
        method="POST",
        url="https://test.test/predictions/7546/post-processings/SurfaceVTP",
        json={"id": "01010101654", "state": "successful"},
        status_code=200,
    )
    for _ in range(2):
        # Mock request for PP data
        httpx_mock.add_response(
            method="GET",
            url="https://test.test/post-processings/01010101654",
            json={
                "data": {"resources": {"vtk": "https://s3.test.test/slice.npz?key=445461321"}},
                "id": "01010101654",
                "location": {},
                "prediction": "7546",
                "type": "SurfaceVTP",
            },
            status_code=200,
        )
    # Mock request for npz binary download
    httpx_mock.add_response(
        method="GET",
        url="https://s3.test.test/slice.npz?key=445461321",
        content=b"this-is-vtp-binary-data",
        headers={"Content-Type": "application/octet-stream"},
        status_code=200,
    )

    pred = simai_client._prediction_directory._model_from({"id": "7546", "state": "successful"})
    surface_vtp = pred.post.surface_vtp()
    assert isinstance(surface_vtp, SurfaceVTP)
    assert isinstance(surface_vtp.data, DownloadableResult)
    # Test a binary download
    binary_io = surface_vtp.data.in_memory()
    assert isinstance(binary_io, BytesIO)
    data_line = binary_io.readline()
    assert data_line.decode("ascii") == "this-is-vtp-binary-data"

    # we have used vtp.data twice, generating 2 hits to the endpoint
    requests = httpx_mock.get_requests()
    assert (
        len([r for r in requests if r.url == "https://test.test/post-processings/01010101654"]) == 2
    )


def test_post_processing_result_surface_vtp_td_location(simai_client, httpx_mock):
    """WHEN Running a surface_vtp post-processing on a prediction
    WITH predict-as-learnt option (location is PPSurfaceLocation.AS_LEARNT)
    AND calling its .data field,
    THEN a GET request is made on the post-processings/SurfaceVTPTDLocation endpoint
    AND the returned instance is of type SurfaceVTPTDLocation.
    """
    # Mock request for PP creation
    httpx_mock.add_response(
        method="POST",
        url="https://test.test/predictions/7546/post-processings/SurfaceVTPTDLocation",
        json={"id": "01010101654", "state": "successful"},
        status_code=200,
    )

    pred = simai_client._prediction_directory._model_from({"id": "7546", "state": "successful"})
    surface_vtp = pred.post.surface_vtp_td_location()
    requests = httpx_mock.get_requests()
    assert (
        len(
            [
                r
                for r in requests
                if r.url
                == "https://test.test/predictions/7546/post-processings/SurfaceVTPTDLocation"
            ]
        )
        == 1
    )
    assert isinstance(surface_vtp, SurfaceVTPTDLocation)


def test_post_processing_result_point_cloud(simai_client, httpx_mock):
    httpx_mock.add_response(
        method="POST",
        url="https://test.test/predictions/7546/post-processings/CustomVolumePointCloud",
        json={"id": "01010101654", "state": "successful"},
        status_code=200,
    )
    for _ in range(2):
        # Mock request for PP data
        httpx_mock.add_response(
            method="GET",
            url="https://test.test/post-processings/01010101654",
            json={
                "data": {
                    "resources": {"vtp": "https://s3.test.test/point_cloud.vtp?key=445461321"}
                },
                "id": "01010101654",
                "location": {},
                "prediction": "7546",
                "type": "CustomVolumePointCloud",
            },
            status_code=200,
        )
    httpx_mock.add_response(
        method="GET",
        url="https://s3.test.test/point_cloud.vtp?key=445461321",
        status_code=200,
        content=b"this-is-vtp-binary-data",
    )
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/geometries/4321",
        json={"id": "4321", "point_cloud": {"id": "2345"}},
    )

    pred = simai_client._prediction_directory._model_from(
        {"id": "7546", "state": "successful", "geometry_id": "4321"}
    )
    point_cloud = pred.post.custom_volume_point_cloud()
    assert isinstance(point_cloud, CustomVolumePointCloud)
    assert isinstance(point_cloud.data, DownloadableResult)
    # Test a binary download
    binary_io = point_cloud.data.in_memory()
    assert isinstance(binary_io, BytesIO)
    data_line = binary_io.readline()
    assert data_line.decode("ascii") == "this-is-vtp-binary-data"

    # we have used vtp.data twice, generating 2 hits to the endpoint
    requests = httpx_mock.get_requests()
    assert (
        len([r for r in requests if r.url == "https://test.test/post-processings/01010101654"]) == 2
    )



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
from threading import Event
from typing import NamedTuple
from unittest.mock import Mock

import httpx
import pytest

from ansys.simai.core.api.client import ApiClient
from ansys.simai.core.data.post_processings import GlobalCoefficients, VolumeVTU
from ansys.simai.core.errors import ConnectionError, ProcessingError
from ansys.simai.core.utils.configuration import ClientConfig


@pytest.mark.parametrize("sse_disabled", [True, False])
def test_api_client_connects_to_sse_if_flag(httpx_mock, sse_disabled):
    """WHEN ApiClient is created
    THEN it connects to the SSE URL if and only if no_sse_connection is not False
    """
    did_connect_to_sse_url = Event()
    should_connect = not sse_disabled

    if should_connect:

        def sse_connection(request):
            did_connect_to_sse_url.set()
            return httpx.Response(200, headers={"Content-Type": "text/event-stream"}, text="{}")

        httpx_mock.add_callback(
            sse_connection,
            method="GET",
            url="https://test.test/sessions/events",
            match_headers={"Accept": "text/event-stream"},
        )
    api = ApiClient(
        ClientConfig(
            organization="toto",
            url="https://test.test",
            _disable_authentication=True,
            # We need this flag as unit tests will otherwise be stuck
            # waiting for a second event (if we set the flag after start).
            # Also __del__ cannot be trusted to be called.
            _stop_sse_threads=True,
            no_sse_connection=sse_disabled,
        )
    )
    # wait 1 sec max if a connection has been made
    did_connect = did_connect_to_sse_url.wait(1)
    assert did_connect == should_connect
    api.check_for_sse_error()


def test_api_client_sse_endpoint_unreachable(httpx_mock):
    """WHEN ApiClient is created, if SSE endpoint is not reachable
    THEN a ConnectionError is raised
    """
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/sessions/events",
        text="Not found",
        status_code=404,
    )
    with pytest.raises(ConnectionError):
        ApiClient(
            ClientConfig(
                organization="toto",
                url="https://test.test",
                _disable_authentication=True,
            )
        )


def test_api_client_sse_endpoint_wrong_organization(httpx_mock):
    """WHEN ApiClient is created, if SSE endpoint returns 403
    THEN a ConnectionError is raised with correct error message
    """
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/sessions/events",
        text="User does not belong to the organization not_extrality or the organization does not exist.",
        status_code=403,
    )
    with pytest.raises(
        ConnectionError,
        match="User does not belong to the organization not_extrality or the organization does not exist.",
    ):
        ApiClient(
            ClientConfig(
                organization="not_extrality",
                url="https://test.test",
                _disable_authentication=True,
            )
        )


def test_wait_non_blocking_for_non_loading_items(simai_client, httpx_mock):
    """WHEN Creating Mesh, prediction, post-processing objects that are finished
    THEN wait can be called and are not blocking
    """
    httpx_mock.add_response(
        method="POST",
        url="https://test.test/geometries/7412/predictions",
        json={"id": "2222", "state": "successful"},
        status_code=200,
    )

    httpx_mock.add_response(
        method="POST",
        url="https://test.test/predictions/2222/post-processings/GlobalCoefficients",
        json={"id": "3333", "state": "successful"},
        status_code=200,
    )
    geometry = simai_client.geometries._model_from(
        {"id": "7412", "state": "successful", "predictions": []}
    )
    pred = geometry.run_prediction(Vx=10)
    assert pred.id == "2222"
    global_coefficients = pred.post.global_coefficients()
    assert global_coefficients.id == "3333"
    assert not geometry.is_pending
    assert not pred.is_pending
    assert not global_coefficients.is_pending
    assert geometry.wait() is True
    assert pred.wait() is True
    assert global_coefficients.wait() is True


create_sse_event = NamedTuple("SSEEvent", [("data", dict)])


def test_sse_event_prediction_success(sse_mixin, prediction_factory):
    """WHEN SSEMixin receives an SSE message updating a prediction to success
    THEN the appropriate prediction is updated and changes state to successful
    """
    pred = prediction_factory(state="processing")
    assert pred.is_pending
    # Mock an SSE success event
    updated_record = pred.fields.copy()
    updated_record.update({"state": "successful", "confidence_score": "high"})
    sse_mixin._handle_sse_event(
        create_sse_event(
            f'{{"target": {{"id": "{pred.id}", "type": "prediction"}}, "status": "successful", "type": "job", "record": {json.dumps(updated_record)}}}'
        )
    )
    assert not pred.is_pending
    assert pred.is_ready
    assert pred.confidence_score == "high"


@pytest.mark.parametrize(
    "start_status, status, record",
    [
        ("processing", "successful", {"confidence_score": "yolo"}),
        ("successful", "processing", {}),
    ],
)
def test_prediction_fields_updated_on_sse_event(
    sse_mixin, prediction_factory, start_status, status, record
):
    prediction = prediction_factory(state=start_status)
    updated_record = prediction.fields.copy()
    updated_record.update(record)
    sse_event = {
        "target": {"id": prediction.id, "type": "prediction"},
        "status": status,
        "type": "job",
    }
    if record:
        sse_event["record"] = updated_record

    sse_mixin._handle_sse_event(create_sse_event(json.dumps(sse_event)))

    assert prediction.fields.get("confidence_score") == record.get("confidence_score")


def test_sse_event_update_prediction_failure(sse_mixin, prediction_factory):
    """WHEN SSEMixin receives a SSE message updating a prediction as failed
    THEN the appropriate prediction object is set to failed and failure_reason field is as expected
    """
    pred = prediction_factory(state="processing")
    assert pred.is_pending
    # Mock a SSE failure event
    updated_record = pred.fields.copy()
    updated_record.update({"state": "failure", "error": "something went wrong"})
    sse_mixin._handle_sse_event(
        create_sse_event(
            f'{{"target": {{"id": "{pred.id}", "type": "prediction"}}, "status": "failure", "type": "job", "record": {json.dumps(updated_record)}}}'
        )
    )
    assert not pred.is_pending
    assert pred.has_failed
    assert pred.failure_reason == "something went wrong"
    with pytest.raises(ProcessingError):
        pred.wait()


def test_wait_timeout_false(simai_client, httpx_mock):
    """WHEN wait timed out
    THEN the return value reflects that to the user (is False)
    """
    httpx_mock.add_response(
        method="POST",
        url="https://test.test/geometries/7412/predictions",
        json={"id": "2222", "state": "pending"},
        status_code=200,
    )
    geometry = simai_client.geometries._model_from(
        {"id": "7412", "state": "successful", "predictions": []}
    )
    pred = geometry.run_prediction(Vx=10)
    assert pred.id == "2222"
    assert pred.wait(0.001) is False


def test_global_wait(simai_client):
    """WHEN simai_client's wait() method is called,
    THEN the wait() method is called once for each registered object.
    """
    pred1 = simai_client._prediction_directory._model_from(
        {"id": "m7541d9", "is_complete": False, "state": "processing"}
    )
    pred2 = simai_client._prediction_directory._model_from(
        {"id": "u87e14", "is_complete": True, "state": "successful"}
    )
    pred3 = simai_client._prediction_directory._model_from(
        {"id": "c963f778", "is_complete": True, "state": "failure"}
    )
    pp1 = simai_client._post_processing_directory._model_from(
        pp_class=GlobalCoefficients,
        data={"id": "ch4687x", "is_complete": False, "state": "processing"},
        prediction=pred2,
    )
    pp2 = simai_client._post_processing_directory._model_from(
        pp_class=VolumeVTU,
        data={"id": "pmx37354", "is_complete": True, "state": "successful"},
        prediction=pred2,
    )

    all_models = [pred1, pred2, pred3, pp1, pp2]
    for model in all_models:
        model.wait = Mock()
    simai_client.wait()
    for model in all_models:
        model.wait.assert_called_once()


def test_prediction_data_update(simai_client, prediction_factory):
    """WHEN simai_client's wait() method is called,
    THEN the wait() method is called once for each registered object.
    """
    pred = prediction_factory(state="processing")
    updated_record = pred.fields.copy()
    updated_record.update({"state": "successful", "confidence_score": "low"})
    simai_client._prediction_directory._handle_sse_event(
        {
            "reason": "",
            "status": "successful",
            "result": {"data": {"values": {"confidence_score": "low"}}, "dim": 1},
            "target": {"id": pred.id, "type": "prediction"},
            "type": "job",
            "record": updated_record,
        }
    )
    assert pred.is_ready
    assert pred.confidence_score == "low"



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json

import httpx
import pytest

from ansys.simai.core.data.downloads import DownloadableResult
from ansys.simai.core.data.post_processings import GlobalCoefficients
from ansys.simai.core.data.selection_post_processings import ExportablePPList
from ansys.simai.core.data.selections import Selection


@pytest.fixture()
def selection_factory(simai_client) -> Selection:
    """Returns a function to create a :py:class:`Selection`."""

    def _factory(predictions=None, **kwargs) -> Selection:
        selection = Selection(geometries=[], scalars=[])
        selection.get_predictions = lambda: predictions or []
        return selection

    return _factory


def test_post_processing_export_global_coefficients(simai_client, httpx_mock):
    """WHEN I call export() on a GlobalCoefficients post-processing
    THEN I get a DownloadableResult object allowing me to download the content.
    """

    pp = simai_client._post_processing_directory._model_from(
        {"id": "zebu", "type": "GlobalCoefficients", "state": "successful"}
    )

    def request_callback(request: httpx.Request):
        payload = json.loads(request.content)
        # check export format and id are as expected
        assert payload["format"] == "json"
        assert payload["ids"] == ["zebu"]
        return httpx.Response(200, json={"data": "it's here"})

    httpx_mock.add_callback(
        request_callback,
        method="POST",
        url="https://test.test/post-processings/export",
    )

    res = pp.export()
    assert isinstance(res, DownloadableResult)
    data = res.in_memory().readline()
    assert json.loads(data.decode("ascii")) == {"data": "it's here"}


def test_post_processing_export_surface_evolution_excel(simai_client, httpx_mock):
    """WHEN I call export() on a SurfaceEvolution post-processing
    THEN I get a DownloadableResult object allowing me to download the content.
    """

    pp = simai_client._post_processing_directory._model_from(
        {"id": "mozeu", "type": "SurfaceEvolution", "state": "successful"}
    )

    def request_callback(request: httpx.Request):
        payload = json.loads(request.content)
        # check export format and id are as expected
        assert payload["format"] == "xlsx"
        assert payload["ids"] == ["mozeu"]
        return httpx.Response(200, content=b"some binary excel data")

    httpx_mock.add_callback(
        request_callback,
        method="POST",
        url="https://test.test/post-processings/export",
    )

    res = pp.export(format="xlsx")
    assert isinstance(res, DownloadableResult)
    data = res.in_memory().readline()
    assert data == b"some binary excel data"


def test_post_processing_selection_export(
    selection_factory, prediction_factory, post_processing_factory, httpx_mock
):
    """WHEN I call export() on a selection.post.global_coefficients()
    THEN the post-processing/export is called with the ids of expected post-processings
    ALSO I get a DownloadableResult in return
    """
    pp1 = post_processing_factory(type=GlobalCoefficients._api_name())
    pp2 = post_processing_factory(type=GlobalCoefficients._api_name())
    selection = selection_factory(
        predictions=[
            prediction_factory(
                post_processings=[
                    pp1,
                ]
            ),
            prediction_factory(
                post_processings=[
                    pp2,
                ]
            ),
        ]
    )

    def request_callback(request: httpx.Request):
        payload = json.loads(request.content)
        # check export format and id are as expected
        assert payload["format"] == "xlsx"
        assert payload["ids"] == [pp1.id, pp2.id]
        return httpx.Response(200, content=b"some binary excel data")

    httpx_mock.add_callback(
        request_callback,
        method="POST",
        url="https://test.test/post-processings/export",
    )
    post_processings = selection.post.global_coefficients()
    assert isinstance(post_processings, ExportablePPList)
    res = post_processings.export(format="xlsx")
    assert isinstance(res, DownloadableResult)
    data = res.in_memory().readline()
    assert data == b"some binary excel data"



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
.. _ref_subset_assignment:

Subset assignment
=================

This example demonstrates how to assign a subset
to a training data.

"""

###############################################################################
# Import necessary libraries
# --------------------------

import ansys.simai.core as asc

simai_client = asc.from_config()

###############################################################################
# Select a training data
# ----------------------
# Example of a training_data_id associated with a project_id.

training_data_id = "k4z77qzq"
project_id = "k9756vw0"

###############################################################################
# Get subset assignment
# ---------------------
# Get and print the current subset assigned for this training_data_id.

current_subset = simai_client.training_data.get(id=training_data_id).get_subset(project=project_id)
print(current_subset)

###############################################################################
# Assign a new subset (two options)
# ---------------------------------
# Manually assign a new subset to the training data.

simai_client.training_data.get(id=training_data_id).assign_subset(project=project_id, subset="Test")

###############################################################################
# Alternatively, use SubsetEnum to assign a valid enum value to the training data.

from ansys.simai.core.data.types import SubsetEnum

simai_client.training_data.get(id=training_data_id).assign_subset(
    project=project_id, subset=SubsetEnum.TEST
)



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""".. _ref_model_configuration_reuse:

Model configuration reuse
=========================

This example demonstrates how to retrieve the latest model configuration
of a project and use it to launch a model build in another project.

"""

###############################################################################
# Import necessary libraries
# --------------------------

import ansys.simai.core as asc

simai_client = asc.from_config()

###############################################################################
# Create a project and allocate training data
# -------------------------------------------
# Define the project name.
new_project_name = "new-project"

###############################################################################
# Create the project.
new_project = simai_client.projects.create(new_project_name)

###############################################################################
# Set the names of the data samples to be associated with the created project.
training_samples_name = [
    "TrainingData_001",
    "TrainingData_002",
    "TrainingData_003",
    "TrainingData_004",
]

###############################################################################
# Retrieve the desired training data samples and associate them with
# the new project.
for td_name in training_samples_name:
    filt = {"name": td_name}
    td = simai_client.training_data.list(filters=filt)
    td[0].add_to_project(new_project)

###############################################################################
# Select a model configuration and associate it with the newly created project
# ----------------------------------------------------------------------------
# Retrieve the model configuration from another project that you wish to reuse.
old_project = "old-ps"
my_project = simai_client.projects.get(name=old_project)

last_build_config = my_project.last_model_configuration

###############################################################################
# If the new project meets the requirements for training, associate
# the project's ID with the configuration and launch a model build.
if new_project.is_trainable():
    # Assign the new project's ID to the configuration to transfer the
    # configuration from the old project to the new one
    last_build_config.project_id = new_project.id

    # Launch a model build for the new project
    new_model = simai_client.models.build(last_build_config)



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
from typing import TYPE_CHECKING, BinaryIO, List, Optional, Union

from ansys.simai.core.data.base import DataModel, Directory
from ansys.simai.core.data.geomai.models import GeomAIModelConfiguration
from ansys.simai.core.data.types import File, Identifiable, get_id_from_identifiable

if TYPE_CHECKING:
    from ansys.simai.core.data.geomai.predictions import GeomAIPrediction
    from ansys.simai.core.data.geomai.projects import GeomAIProject


class GeomAIWorkspace(DataModel):
    """Provides the local representation of a GeomAI workspace object."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._model_manifest = None
        self._model_configuration = None

    def __repr__(self) -> str:
        return f"<GeomAIWorkspace: {self.id}, {self.name}>"

    @property
    def name(self) -> str:
        """Name of the workspace."""
        return self.fields["name"]

    @property
    def model_configuration(self) -> GeomAIModelConfiguration:
        """Model configuration used in the workspace."""
        if self._model_configuration is None:
            model_config = self._client._api.get_workspace_model_configuration(self.id)
            self._model_configuration = GeomAIModelConfiguration(**model_config)
        return self._model_configuration

    def rename(self, new_name: str) -> None:
        """Rename the workspace.

        Args:
            new_name: New name to give to the workspace.
        """
        self._client._api.update_geomai_workspace(self.id, name=new_name)
        self.reload()

    def delete(self):
        """Delete the workspace."""
        return self._client._api.delete_geomai_workspace(self.id)

    def list_predictions(self) -> List["GeomAIPrediction"]:
        """Lists all the predictions in the workspace."""
        return [
            self._client.geomai.predictions._model_from(prediction)
            for prediction in self._client._api.geomai_predictions(self.id)
        ]

    def set_as_current_workspace(self) -> None:
        """Configure the client to use this workspace instead of the one currently configured."""
        self._client.geomai.current_workspace = self

    def download_latent_parameters_json(self, file: Optional[File] = None) -> Union[None, BinaryIO]:
        """Download the json file containing the latent parameters for the model's training data.

        Args:
            file: Binary file-object or the path of the file to put the content into.

        Returns:
            ``None`` if a file is specified or a binary file-object otherwise.
        """
        return self._client._api.download_geomai_workspace_latent_parameters(self.id, file)

    def get_latent_parameters(self) -> dict[str, List[float]]:
        """Get the dictionary mapping geometry names to their latent parameter vectors for the model's training data."""
        data = self._client._api.download_geomai_workspace_latent_parameters(self.id, None)
        if data is None:
            return {}
        latent_parameters = json.loads(data.read().decode("utf-8"))
        return latent_parameters

    def download_model_evaluation_report(
        self, file: Optional[File] = None
    ) -> Union[None, BinaryIO]:
        """Download the ZIP file of the model evaluation report for the workspace.

        Args:
            file: Binary file-object or the path of the file to put the content into.

        Returns:
            ``None`` if a file is specified or a binary file-object otherwise.
        """
        return self._client._api.download_geomai_workspace_model_evaluation_report(self.id, file)


class GeomAIWorkspaceDirectory(Directory[GeomAIWorkspace]):
    """Provides a collection of methods related to GeomAI workspaces.

    This class is accessed through ``client.geomai.workspaces``.

    Example:
      .. code-block:: python

            import ansys.simai.core as asc

            simai_client = asc.from_config()
            simai_client.geomai.workspaces.list()
    """

    _data_model = GeomAIWorkspace

    def list(self) -> List[GeomAIWorkspace]:
        """List all workspaces from the server."""
        return [self._model_from(workspace) for workspace in self._client._api.geomai_workspaces()]

    def get(self, id: Optional[str] = None, name: Optional[str] = None) -> GeomAIWorkspace:
        """Get a specific workspace object from the server by either ID or name.

        You can specify either the ID or the name, not both.

        Args:
            id: ID of the workspace.
            name: Name of the workspace.

        Returns:
            A :class:`GeomAIWorkspace`.

        Raises:
            NotFoundError: No workspace with the given ID exists.
        """
        if name and id:
            raise ValueError("'id' and 'name' cannot both be specified.")
        if name:
            return self._model_from(self._client._api.get_geomai_workspace_by_name(name))
        if id:
            return self._model_from(self._client._api.get_geomai_workspace(id))
        raise ValueError("Either 'id' or 'name' must be specified.")

    def create(self, name: str, project: Identifiable["GeomAIProject"]) -> GeomAIWorkspace:
        """Create a workspace.

        Args:
            name: Name to give the new workspace.
            project: ID or :class:`project <.projects.GeomAIProject>` of the workspace.
        """
        return self._model_from(
            self._client._api.create_geomai_workspace(name, get_id_from_identifiable(project))
        )

    def delete(self, workspace: Identifiable[GeomAIWorkspace]) -> None:
        """Delete a workspace.

        Args:
            workspace: ID or :class:`model <GeomAIWorkspace>` of the workspace.
        """
        self._client._api.delete_geomai_workspace(get_id_from_identifiable(workspace))
